import pyproj


def write_projection(dset, config):
    crs = pyproj.CRS.from_user_input(config['proj4'])
    attrs = crs.to_cf()
    if 'horizontal_datum_name' not in attrs:
        attrs['horizontal_datum_name'] = 'World Geodetic System 1984'
    dset.createVariable('crs', data=0, dims=(), attrs=attrs)
    cs = crs.cs_to_cf()
    dset.setAttrs(config['x'], cs[0])
    dset.setAttrs(config['y'], cs[1])
    dset.setAttrs(config['output_varname'], dict(grid_mapping='crs'))
    dset.main_dataset.Conventions = "CF-1.8"



class Example:
    def __init__(self, name):
        self.name = name
        self._files = None
        self._config = None

    @staticmethod
    def available():
        import pkgutil
        return [m.name for m in pkgutil.iter_modules(__path__) if m.ispkg]

    @property
    def package(self):
        return __name__ + '.' + self.name

    @property
    def config(self):
        if self._config is None:
            import yaml
            data = self.files()['aggregate.yaml']
            self._config = yaml.safe_load(data.decode('utf-8'))
        return self._config

    def files(self, glob=None):
        if glob is None:
            return self.data

        import re
        re_pattern = '^' + glob.replace('.', '\\.').replace('?', '.').replace('*', '.*') + '$'
        return {k: v for k, v in self.data.items() if re.match(re_pattern, k)}

    @property
    def descr(self):
        import re
        import pkgutil
        data = pkgutil.get_data(self.package, 'aggregate.yaml')
        config_txt = data.decode(encoding='utf-8')
        match = re.match('^# (.*)', config_txt)
        if match:
            return match.group(1)
        else:
            return ""

    @property
    def data(self):
        if self._files is None:
            self._files = self._load()
        return self._files

    def _load(self):
        import importlib.resources
        names = [
            f for f in importlib.resources.contents(self.package)
            if not f.startswith('__')
        ]

        data = dict()

        for fname in names:
            import pkgutil
            data[fname] = pkgutil.get_data(self.package, fname)

            if fname.endswith('.nc.yaml'):
                import yaml
                import xarray as xr
                key = fname[:-5]
                yaml_text = data[fname].decode(encoding='utf-8')
                xr_dict = yaml.safe_load(yaml_text)
                xr_dset = xr.Dataset.from_dict(xr_dict)
                data[key] = xr_dset

        return data

    def extract(self, dirname='.'):
        import logging
        from pathlib import Path
        import xarray as xr
        logger = logging.getLogger(__name__)
        outdir = Path(dirname)

        output_files = [k for k in self.files(self.config['outfile'])]
        input_files = {k: v for k, v in self.files().items() if k not in output_files}

        for fname, data in input_files.items():
            if fname.endswith('.nc.yaml'):
                continue

            logger.info(f'Extract input file: "{fname}"')
            path = outdir / fname
            if fname.endswith('.nc'):
                assert isinstance(data, xr.Dataset)
                data.to_netcdf(path)
            else:
                with open(path, 'bw') as f:
                    f.write(data)

        return 'aggregate.yaml'

    def run(self):
        from .. import script
        from ..output import MultiDataset
        from ..input import LadimInputStream

        ladim_dsets = list(self.files(self.config['infile']).values())
        ladim_input_stream = LadimInputStream(ladim_dsets)

        outfile_name = self.config['outfile']
        with MultiDataset(outfile_name, diskless=True) as output_dset:
            script.run(ladim_input_stream, self.config, output_dset, self.files())
            result = output_dset.to_dict()

        import yaml
        out_pattern = self.config['outfile'].replace('.nc', '*.nc.yaml')
        expected = {
            k[:-5]: yaml.safe_load(v.decode(encoding='utf-8'))
            for k, v in self.files(out_pattern).items()
        }

        return result, expected


def nc_dump(dset):
    """Returns the contents of an open netCDF4 dataset as a dict"""

    variables = dict()
    for name in dset.variables:
        v = dict()
        v['dims'] = list(dset.variables[name].dimensions)
        if len(v['dims']) == 1:
            v['dims'] = v['dims'][0]

        v['data'] = dset.variables[name][:].tolist()

        atts = dict()
        for attname in dset.variables[name].ncattrs():
            atts[attname] = dset.variables[name].getncattr(attname)
        if atts:
            v['attrs'] = atts

        variables[name] = v

    return variables



import netCDF4 as nc
import numpy as np
import logging


logger = logging.getLogger(__name__)


class MultiDataset:
    """Encapsulates a collection of netCDF4 datasets.

    The class contains a main dataset which holds all the coordinate variables.
    Some of the coordinate variables are defined as "cross-dataset"
    coordinates. Variables having any of the cross-dataset dimensions are
    spread across multiple datasets. Variables with only "regular" coordinates
    are kept in the main dataset.
    """

    def __init__(self, filename, **kwargs):
        """
        Initialize a MultiDataset. The file of the main dataset is given
        explicitly. The names of the subdatasets are constructed from the main
        filename, of the form "maindatasetname_firstcoordvalue_secondcoordvalue".

        :param filename: Name of the main dataset.
        :param kwargs: Additional arguments passed to netCDF4.Dataset
        """
        self._dataset_kwargs = kwargs
        self.datasets = dict()
        self._cross_coords = dict()
        self._cross_vars = dict()
        self._editable = True  # When subdatasets are created, no more variables can be added

        if isinstance(filename, nc.Dataset):
            self.main_dataset = filename
            self._has_diskless_input_dataset = True
            self._dataset_kwargs['diskless'] = True
        else:
            self.main_dataset = nc.Dataset(filename, mode='w', **kwargs)
            self._has_diskless_input_dataset = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        logger.info(f'Close dataset "{self.main_dataset.filepath()}"')
        if not self._has_diskless_input_dataset:
            self.main_dataset.close()
        for d in self.datasets.values():
            logger.info(f'Close dataset "{d.filepath()}"')
            d.close()

    def createCoord(self, varname, data, attrs=None, cross_dataset=False):
        """
        Create coordinate variable and dimension in the main dataset.

        Coordinate variables cannot be created after setData or getData is used
        on a cross-dataset variable.

        :param varname: Name of coordinate variable and dimension
        :param data: Data used to initialize variable
        :param attrs: Optional coordinate attributes
        :param cross_dataset: True if this is a cross-dataset dimension
        :return: The newly created variable
        """
        if not self._editable:
            raise TypeError(
                "Adding new variables must be done before accessing cross-dataset data")

        data = np.array(data)

        if np.issubdtype(data.dtype, np.datetime64):
            attrs = {
                **(attrs or dict()),
                **dict(
                    units="microseconds since 1970-01-01",
                    calendar="proleptic_gregorian",
                ),
            }
            datediff = data - np.datetime64('1970-01-01')
            data = datediff.astype('timedelta64[us]').astype('i8')

        dset = self.main_dataset
        dset.createDimension(varname, len(data))
        variable = dset.createVariable(varname, data.dtype, varname, fill_value=False)
        variable.set_auto_maskandscale(False)
        variable[:] = data
        if attrs:
            dset.variables[varname].setncatts(attrs)
        if cross_dataset:
            self._cross_coords[varname] = data
        return variable

    def setAttrs(self, varname, attrs):
        """
        Set attributes of a regular variable
        :param varname:
        :param attrs:
        :return:
        """
        if not self._editable:
            # We cannot edit attributes of any variables. If we were to allow this, we
            # would need to change attributes of all sub-datasets as well.
            raise TypeError(
                "Setting attributes must be done before accessing cross-dataset data")

        # Set cross-dataset variable attributes
        if varname in self._cross_vars:
            old_attrs = self._cross_vars[varname]['attrs']
            self._cross_vars[varname]['attrs'] = {**old_attrs, **attrs}

        # Set regular variable attributes
        else:
            v = self.main_dataset.variables[varname]
            v = v  # type: nc.Variable
            v.setncatts(attrs)

    def getCoord(self, varname):
        """
        Return a coordinate variable from the main dataset
        :param varname: Variable name
        :return: The variable object
        """
        return self.main_dataset.variables[varname]

    def getAttrs(self, varname):
        """
        Return the attributes of either a regular or cross-dataset
        variable.

        :param varname: Variable name
        :return: A dict of attributes
        """

        # Get cross-dataset variable attributes
        if varname in self._cross_vars:
            return self._cross_vars[varname]['attrs']

        # Get regular variable attributes
        else:
            v = self.main_dataset.variables[varname]
            return {k: v.getncattr(k) for k in v.ncattrs()}

    def createVariable(self, varname, data, dims, attrs=None):
        """
        Create a dataset variable. If any of the dimensions are cross-dataset,
        the variable will be spread out over multiple datasets.

        Variables cannot be created after setData or getData is used on a
        cross-dataset variable.

        :param varname: Variable name
        :param data: The initial data of the variable
        :param dims: A tuple of dimension names
        :param attrs: An optional dict of attributes
        :return: The variable object if regular variable, None otherwise
        """
        if not self._editable:
            raise TypeError(
                "Adding new variables must be done before accessing cross-dataset data")

        data = np.array(data)

        # A single dimension should be converted to list
        if isinstance(dims, str):
            dims = [dims]

        # Create cross-dataset variable
        if any(d in self._cross_coords for d in dims):
            self._cross_vars[varname] = dict(
                data=data,
                dims=dims,
                attrs=attrs or dict(),
            )
            variable = None

        # Create regular variable
        else:
            dset = self.main_dataset
            variable = dset.createVariable(varname, data.dtype, dims, fill_value=False)
            variable.set_auto_maskandscale(False)
            variable[:] = data
            if attrs:
                variable.setncatts(attrs)

        return variable

    def _get_filename(self, file_idx):
        import os

        # Get filename-friendly value of variable
        def strval(varname, index):
            variable = self.getCoord(varname)
            attrs = self.getAttrs(varname)
            value = variable[index]
            if 'since' in attrs.get('units', ''):
                import cftime
                date = cftime.num2date(value, attrs['units'],
                                       attrs.get('calendar', 'standard'))
                return f'{date:%Y%m%d%H%M%S}'
            else:
                return str(value)

        stubs = [str(strval(k, v)) for k, v in file_idx.items()]
        base, ext = os.path.splitext(self.main_dataset.filepath())
        return base + '_' + '_'.join(stubs) + ext

    def getDataset(self, **file_idx):
        """Return sub-dataset, or create it if necessary"""
        file_idx_key = tuple((k, file_idx[k]) for k in self._cross_coords.keys())
        if file_idx_key not in self.datasets:
            fname = self._get_filename(file_idx)
            logger.info(f'Create sub-dataset "{fname}"')
            dset = nc.Dataset(fname, mode='w', **self._dataset_kwargs)
            self._copy_from_main(dset, file_idx)
            self.datasets[file_idx_key] = dset
            self._editable = False

        return self.datasets[file_idx_key]

    def _copy_from_main(self, dest, crossdim_idx):
        cross_dims = list(self._cross_coords.keys())
        source = self.main_dataset

        # Copy dimensions
        for dim in source.dimensions.values():
            if dim.name in cross_dims:
                dest.createDimension(dim.name, 1)
            else:
                dest.createDimension(dim.name, dim.size)

        # Copy variables
        for v in source.variables.values():
            new_var = dest.createVariable(v.name, v.dtype, v.dimensions)
            new_var.set_auto_maskandscale(False)
            atts = {k: v.getncattr(k) for k in v.ncattrs()}
            new_var.setncatts(atts)
            if v.name in cross_dims:
                # If crossdim coordinate, copy only one value
                global_idx = crossdim_idx[v.name]
                new_var[:] = v[global_idx]
                new_var.setncattr('global_index', global_idx)
            else:
                # Otherwise, copy all data
                new_var[:] = v[:]

        # Add filesplit variables
        for vname, vinfo in self._cross_vars.items():
            new_var = dest.createVariable(vname, vinfo['data'].dtype, vinfo['dims'])
            new_var.set_auto_maskandscale(False)
            new_var[:] = vinfo['data']
            dest.variables[vname].setncatts(vinfo['attrs'])

        # Copy global attributes
        atts = {k: source.getncattr(k) for k in source.ncattrs()}
        dest.setncatts(atts)

    def _get_shape_of_variable_data(self, varname, idx):
        shape = []
        for dim_slice, dim_name in zip(idx, self._cross_vars[varname]['dims']):
            dim_size = self.main_dataset.dimensions[dim_name].size
            dim_range = range(*dim_slice.indices(dim_size))
            shape.append(len(dim_range))
        return tuple(shape)

    def getData(self, varname, idx=None):
        """
        Get data from a dataset variable.

        :param varname: Variable name
        :param idx: A tuple of slices, or None if the entire data range is desired.
        :return: A numpy array of data
        """

        if varname in self._cross_vars:
            variable_info = self._cross_vars[varname]
            if idx is None:
                idx = (slice(None),) * len(variable_info['dims'])

            shape = self._get_shape_of_variable_data(varname, idx)
            data = np.empty(shape, variable_info['data'].dtype)

            for i_file, i_var, i_data in self._split_indices(varname, idx):
                dset = self.getDataset(**i_file)
                data[i_data] = dset.variables[varname][i_var]

            return data

        else:
            if idx is None:
                idx = slice(None)
            return self.main_dataset.variables[varname][idx]

    def setData(self, varname, data, idx=None):
        """
        Set data of a dataset variable

        :param varname: Variable name
        :param data: A numpy array of data
        :param idx: A tuple of slices, or None if the entire data range is desired.
        :return: The input data
        """
        if varname in self._cross_vars:
            variable_info = self._cross_vars[varname]
            if idx is None:
                idx = (slice(None),) * len(variable_info['dims'])

            shape = self._get_shape_of_variable_data(varname, idx)
            bdata = np.broadcast_to(data, shape)

            for i_file, i_var, i_data in self._split_indices(varname, idx):
                dset = self.getDataset(**i_file)

                dset.variables[varname][i_var] = bdata[i_data]

            return data

        else:
            if idx is None:
                idx = slice(None)
            self.main_dataset.variables[varname][idx] = data
        return data

    def incrementData(self, varname, data, idx=None):
        """
        Increment a dataset variable

        :param varname: Variable name
        :param data: A numpy array of data
        :param idx: A tuple of slices, or None if the entire data range is desired.
        :return: The input data
        """
        previous = self.getData(varname, idx)
        newdata = previous + np.broadcast_to(data, previous.shape)
        self.setData(varname, newdata, idx)

    def _split_indices(self, varname, index):
        """
        Iterate over subdatasets to aid access of cross-dataset variables.

        The input index is a tuple of slices, relative to the global shape
        of the variable. The function iterates over each dataset spanned by
        the range.

        For each iteration, returns a tuple `(cross_idx, var_idx, data_idx)`
        representing the portion of the data spanned by a single dataset.
        - `cross_idx` is a dict of the form `{crossdimA: intA, crossdimB: intB, ...}`,
          identifying the subdataset.
        - `var_idx` is a tuple of either zeros or slices. For each
          cross-dataset dimension there is a zero, and for each main dimension
          there is a slice (taken directly from `index`). This tuple can be
          used to directly index variables in a subdataset.
        - `data_idx` is a tuple of slice(None) and integers. For each cross-
          dataset dimension there is an integer, and for each main dimension
          there is slice(None). The integer is relative to the shape of the
          data subset.
        :param varname: The variable name
        :param index: A tuple of slices
        :return: A tuple `(cross_idx, var_idx, data_idx)` representing a
        portion of the data spanned by a single subdataset.
        """

        from itertools import product

        varinfo = self._cross_vars[varname]
        dims = varinfo['dims']
        is_splitdim = [dim in self._cross_coords for dim in dims]
        big_shape = [self.main_dataset.dimensions[d].size for d in dims]
        big_range = [range(*i.indices(s)) for i, s in zip(index, big_shape)]
        shape = [len(range(*i.indices(s))) for i, s in zip(index, big_shape)]

        # Construct the var_idx, which is the index of the subdataset variable
        var_idx = []
        for dim, index_element, sd in zip(dims, index, is_splitdim):
            if sd:
                var_idx.append(0)
            else:
                var_idx.append(index_element)

        # Construct the index elements relative to the big data array
        data_indices = []
        for index_element, dim, sz, sd in zip(index, dims, shape, is_splitdim):
            if sd:
                data_indices.append(range(sz))
            else:
                data_indices.append([slice(None)])

        # Iterate over the cartesian product
        for data_idx in product(*data_indices):
            file_idx = {d: br[di] for d, br, di, sd in
                        zip(dims, big_range, data_idx, is_splitdim) if sd}
            yield file_idx, var_idx, data_idx

    def to_dict(self):
        all_datasets = [self.main_dataset] + list(self.datasets.values())
        dset_dict = {}
        for dset in all_datasets:
            key = dset.filepath()
            dset_dict[key] = nc_to_dict(dset)
        return dset_dict


def nc_to_dict(dset):
    d = dict()
    for varname in dset.variables:
        v = dset.variables[varname]
        d[varname] = dict(
            dims=list(v.dimensions),
            data=v[:].tolist(),
        )
        if len(d[varname]['dims']) == 1:
            d[varname]['dims'] = d[varname]['dims'][0]

        atts = dict()
        for attname in v.ncattrs():
            atts[attname] = v.getncattr(attname)
        if atts:
            d[varname]['attrs'] = atts
    return d



import numpy as np
import pandas as pd
import logging
import cftime
import xarray as xr
import datetime


logger = logging.getLogger(__name__)


class Histogrammer:
    def __init__(self, bins=None):
        self.weights = dict(bincount=None)
        self.coords = Histogrammer._get_coords_from_bins(bins)

    @staticmethod
    def _get_coords_from_bins(bins_dict):
        crd = dict()
        for crd_name, bins in bins_dict.items():
            edges = bins['edges']
            centers = bins['centers']
            attrs = bins.get('attrs', dict())
            crd[crd_name] = dict(centers=centers, edges=edges, attrs=attrs)
        return crd

    def make(self, chunk):
        coord_names = list(self.coords.keys())
        bins = [self.coords[k]['edges'] for k in coord_names]
        coords = []
        for k in coord_names:
            logger.debug(f'Load variable "{k}"')
            coords.append(chunk[k].values)

        if '_auto_weights' in chunk.variables:
            weights = chunk['_auto_weights'].values
        else:
            weights = None

        values, idx = adaptive_histogram(coords, bins, weights=weights)
        yield dict(indices=idx, values=values)


def get_centers_from_edges(edges):
    edgediff = edges[1:] - edges[:-1]
    return edges[:-1] + 0.5 * edgediff


def adaptive_histogram(sample, bins, **kwargs):
    """
    Return an adaptive histogram

    For input values `sample` and `bins`, the code snippet

    hist = np.zeros([len(b) - 1 for b in bins])
    hist_chunk, idx = adaptive_histogram(sample, bins, **kwargs)
    hist[idx] = hist_chunk

    gives the same output as

    hist, _ = np.histogramdd(sample, bins, **kwargs)

    :param sample:
    :param bins:
    :param kwargs:
    :return:
    """

    # Cast datetime samples to be comparable with bins
    for i, s in enumerate(sample):
        s_dtype = np.asarray(s).dtype
        if np.issubdtype(s_dtype, np.datetime64):
            sample[i] = s.astype(bins[i].dtype)

    num_entries = next(len(s) for s in sample)
    included = np.ones(num_entries, dtype=bool)

    # Find histogram coordinates of each entry
    binned_sample = []
    for s, b in zip(sample, bins):
        coords = np.searchsorted(b, s, side='right') - 1
        included = included & (0 <= coords) & (coords < len(b) - 1)
        binned_sample.append(coords)

    # Filter out coordinates outside interval
    for i, bs in enumerate(binned_sample):
        binned_sample[i] = bs[included]

    # Abort if there are no points left
    if len(binned_sample[0]) == 0:
        return np.zeros((0, ) * len(sample)), (slice(1, 0), ) * len(sample)

    # Aggregate particles
    df = pd.DataFrame(np.asarray(binned_sample).T)
    df_grouped = df.groupby(list(range(len(bins))))
    if kwargs.get('weights', None) is None:
        df['weights'] = 1
        df_sum = df_grouped.count()
    else:
        df['weights'] = np.asarray(kwargs['weights'])[included]
        df_sum = df_grouped.sum()
    coords = df_sum.index.to_frame().values.T
    vals = df_sum['weights'].values

    # Find min and max bin edges to be used
    idx = [(np.min(c), np.max(c) + 1) for c in coords]
    idx_slice = [slice(start, stop) for start, stop in idx]

    # Densify
    shifted_coords = coords - np.asarray([start for start, _ in idx])[:, np.newaxis]
    shape = [stop - start for start, stop in idx]
    hist_chunk = np.zeros(shape, dtype=vals.dtype)
    hist_chunk[tuple(shifted_coords)] = vals

    return hist_chunk, tuple(idx_slice)


def convert_datebins(spec: dict, dset: xr.Dataset):
    """
    Convert bin specifications containing dates.

    The function takes a bin specification as input, as well as an xarray.Dataset
    where the data variables may potentially contain cfunits date attributes. If one
    of the bin specifications is specified as a date (either date string or python/numpy
    object), and connected to a variable with date attributes, the bin specification is
    converted to a number according to the variable's cfunits date specification.

    The function works for the following types of bin formats:

    1. Pure list - ['2001-01', np.datetime('2001-01-02T12')]
    2. Label format: dict(edges=['2001-01', '2001-02', '2001-03'], labels=['J', 'F'])
    3. Range format: dict(min='2001-01', max='2001-03', step='1 months')

    Date steps can be specified as 'years', 'months', 'days', 'hours', 'minutes',
    'seconds', or a numpy timedelta64 data type.

    :param spec: A dict where the keys are the variable names and the values are the bin
                 specifications.
    :param dset: A dataset where the variables may have cfunits date attributes
    :return:     A dict of the same format as ``spec``, except that dates are converted
                 to numbers.
    """
    newspec = {}

    # k is the variable name
    # v is the bin format
    for k, v in spec.items():
        if k not in dset:
            attrs = {}
        else:
            attrs = dset[k].attrs
        units = attrs.get('units', '')
        calendar = attrs.get('calendar', 'standard')
        newspec[k] = convert_binspec(v, units, calendar)

    return newspec


def convert_binspec(singlespec, units, calendar):
    if 'since' not in units:
        return singlespec

    elif isinstance(singlespec, list):
        return [convert_date(v, units, calendar) for v in singlespec]

    elif isinstance(singlespec, dict):
        result = {k: v for k, v in singlespec.items()}

        if 'min' in result:
            result['min'] = convert_date(result['min'], units, calendar)

        if 'max' in result:
            result['max'] = convert_date(result['max'], units, calendar)

        if 'edges' in result:
            result['edges'] = [convert_date(e, units, calendar) for e in result['edges']]

        if 'step' in result:
            result['step'] = convert_step(result['step'], units, calendar)

        return result

    elif isinstance(singlespec, str) and singlespec.count(' ') == 1:
        return convert_step(singlespec, units, calendar)

    else:
        return singlespec


def convert_step(step_spec, units, calendar):
    if isinstance(step_spec, str):
        step_str, units_in = step_spec.split(sep=' ', maxsplit=1)
        step = float(step_str)
        units_out, _, refdate = units.split(sep=' ', maxsplit=2)
    else:
        return step_spec

    date_in = cftime.num2date(step, f'{units_in} since {refdate}', calendar)
    return cftime.date2num(date_in, f'{units_out} since {refdate}', calendar)


def convert_date(value, units, calendar):
    """
    Convert the input value to a cfunits date format. The input value can be anything
    convertible to np.datetime64.

    :param value: Input date
    :param units: Output cfunits format
    :param calendar: Output cfunits calendar
    :return: Date in cfunits format
    """
    try:
        value = np.datetime64(value).astype(object)
    except ValueError:
        return value

    if isinstance(value, datetime.date):
        value = datetime.datetime.combine(value, datetime.time())

    return cftime.date2num(value, units, calendar)


def autobins(spec, dset):
    if dset is not None and hasattr(dset, 'open_dataset'):
        with dset.open_dataset(0) as xr_dset:
            spec = convert_datebins(spec, xr_dset)

    # Add INIT bins, if any
    for k in spec:
        if k.endswith('_INIT'):
            varname, opname = k.rsplit(sep='_', maxsplit=1)
            dset.add_init_variable(varname)

    # Find bin specification type
    spec_types = dict()
    for k, v in spec.items():
        if isinstance(v, list):
            spec_types[k] = 'edges'

        elif isinstance(v, dict) and all(u in v for u in ['min', 'max', 'step']):
            spec_types[k] = 'range'

        elif isinstance(v, dict) and all(u in v for u in ['edges', 'labels']):
            spec_types[k] = 'edges_labels'

        elif v == 'group_by' or v == 'unique':
            spec_types[k] = 'unique'

        elif np.issubdtype(type(v), np.number):
            spec_types[k] = 'resolution'

        elif isinstance(v, str):
            specsplit = v.split(sep=" ")
            if len(specsplit) == 2:  # Single number with units
                spec_types[k] = 'resolution'
            else:
                raise ValueError(f'Unknown bin type: {v}')

        else:
            raise ValueError(f'Unknown bin type: {v}')

    # Check if we need pre-scanning of the dataset
    scan_params_template = dict(unique=['unique'], resolution=['min', 'max'])
    scan_params = {k: scan_params_template[v] for k, v in spec_types.items()
                   if v in scan_params_template}
    scan_output = {k: dict() for k in spec}
    if scan_params:
        # First add all the variable definitions...
        aggvars = []
        for varname, aggfuncs in scan_params.items():
            for aggfun in aggfuncs:
                if varname.endswith('_INIT'):
                    # If init variable, use the base variable for aggregation
                    aggvar = dset.add_aggregation_variable(varname[:-5], aggfun)
                else:
                    aggvar = dset.add_aggregation_variable(varname, aggfun)
                aggvars.append((aggvar, varname, aggfun))

        logger.info(f'Scan input dataset to find {", ".join([s[0] for s in aggvars])}')

        # ... then trigger the scanning of the dataset
        for aggvar, varname, aggfun in aggvars:
            scan_output[varname][aggfun] = dset.get_aggregation_value(aggvar)

    # Put the specs and the result of the pre-scanning into the bin generator
    bins = {k: bin_generator(spec[k], spec_types[k], scan_output[k]) for k in spec}

    # Add attributes from the dataset
    if dset is not None and hasattr(dset, 'attributes'):
        for k, v in dset.attributes.items():
            if k in bins:
                bins[k]['attrs'] = v

    return bins


def bin_generator(spec, spec_type, scan_output):
    if spec_type == 'edges':
        edges = np.asarray(spec)
        centers = get_centers_from_edges(edges)
    elif spec_type == 'edges_labels':
        edges = np.asarray(spec['edges'])
        centers = np.asarray(spec['labels'])
    elif spec_type == 'range':
        edges = np.arange(spec['min'], spec['max'] + spec['step'], spec['step'])
        centers = get_centers_from_edges(edges)
    elif spec_type == 'unique':
        data = scan_output['unique']
        edges = np.concatenate([data, [data[-1] + 1]])
        centers = np.asarray(data)
    elif spec_type == 'resolution':
        minval = align_to_resolution(scan_output['min'], spec)
        maxval = align_to_resolution(scan_output['max'] + 2 * spec, spec)
        edges = np.arange(minval, maxval, spec)
        centers = get_centers_from_edges(edges)
    else:
        raise ValueError(f'Unknown spec_type: {spec_type}')

    return dict(edges=edges, centers=centers)


def t64conv(timedelta_or_other):
    """
    Convert input data in the form of [value, unit] to timedelta64, or returns the
    argument verbatim if there are any errors. Also accepts string values of the
    form used in cfunits (e.g., "23 hours")
    """
    if isinstance(timedelta_or_other, str) and timedelta_or_other.count(' ') == 1:
        t64val_str, t64unit_cf = timedelta_or_other.split(sep=" ")
        t64val = int(t64val_str)
        if t64unit_cf.endswith('s') and len(t64unit_cf) > 3:  # Remove plurals
            t64unit_cf = t64unit_cf[:-1]
        unitconv = dict(day='D', hour='h', minute='m', second='s', millisecond='ms',
                        microsecond='us')
        t64unit = unitconv.get(t64unit_cf, t64unit_cf)
        return np.timedelta64(t64val, t64unit)

    try:
        t64val, t64unit = timedelta_or_other
        return np.timedelta64(t64val, t64unit)
    except TypeError:
        return timedelta_or_other


# Align to wholenumber resolution
def align_to_resolution(value, resolution):
    """
    Round down to a specified resolution.

    Specifically, `(returned_value <= value) and (returned_value % resolution == 0)`.
    """
    if np.issubdtype(np.array(resolution).dtype, np.timedelta64):
        val_posix = (value - np.datetime64('1970-01-01')).astype('timedelta64[us]')
        res_posix = resolution.astype('timedelta64[us]')
        ret_posix = (val_posix.astype('i8') // res_posix.astype('i8')) * res_posix
        return np.datetime64('1970-01-01') + ret_posix
    else:
        return np.array((value // resolution) * resolution).item()



from ladim_aggregate import proj
from ladim_aggregate.output import MultiDataset
import pytest
from uuid import uuid4


class Test_write_projection:
    @pytest.fixture()
    def nc_dset(self):
        with MultiDataset(filename=uuid4(), diskless=True) as dset:
            dset.createCoord('X', data=[1, 2, 3, 4])
            dset.createCoord('Y', data=[10, 20, 30])
            dset.createVariable('histogram', data=0, dims=('Y', 'X'))
            yield dset

    def test_adds_crs_variable(self, nc_dset):
        config = dict(proj4="+proj=longlat +ellps=WGS84 +datum=WGS84", x='X', y='Y', output_varname='histogram')
        proj.write_projection(nc_dset, config)
        assert nc_dset.getAttrs('crs')['grid_mapping_name'] == 'latitude_longitude'

    def test_adds_attrs_to_coord_vars(self, nc_dset):
        config = dict(proj4="+proj=longlat +ellps=WGS84 +datum=WGS84", x='X', y='Y', output_varname='histogram')
        proj.write_projection(nc_dset, config)
        assert nc_dset.getAttrs('X')['standard_name'] == 'longitude'
        assert nc_dset.getAttrs('Y')['standard_name'] == 'latitude'

    def test_adds_grid_mapping_to_histogram_var(self, nc_dset):
        config = dict(proj4="+proj=longlat +ellps=WGS84 +datum=WGS84", x='X', y='Y', output_varname='histogram')
        proj.write_projection(nc_dset, config)
        assert nc_dset.getAttrs('histogram')['grid_mapping'] == 'crs'

    def test_adds_conventions_attribute_to_dataset(self, nc_dset):
        config = dict(proj4="+proj=longlat +ellps=WGS84 +datum=WGS84", x='X', y='Y', output_varname='histogram')
        proj.write_projection(nc_dset, config)
        assert nc_dset.main_dataset.Conventions == "CF-1.8"



import netCDF4
import xarray as xr
from ladim_aggregate import output
import pytest
import numpy as np
from uuid import uuid4


class Test_MultiDataset:
    @pytest.fixture(scope="function")
    def mdset(self):
        with output.MultiDataset(uuid4(), diskless=True) as d:
            yield d

    def test_can_set_main_coordinate(self, mdset):
        mdset.createCoord('mycoord', [1, 2, 3])
        assert mdset.getCoord('mycoord')[:].tolist() == [1, 2, 3]

    def test_can_set_main_variable(self, mdset):
        mdset.createCoord('A', [1, 2, 3])
        mdset.createCoord('B', [10, 20])
        mdset.createVariable('myvar', np.arange(6).reshape((2, 3)), ('B', 'A'))
        assert mdset.getData('myvar').tolist() == [[0, 1, 2], [3, 4, 5]]

    def test_can_get_partial_data_of_main_variable(self, mdset):
        mdset.createCoord('A', [1, 2, 3])
        mdset.createCoord('B', [10, 20])
        mdset.createVariable('myvar', np.arange(6).reshape((2, 3)), ('B', 'A'))

        data = mdset.getData('myvar')
        idx = np.s_[:1, 1:]
        assert mdset.getData('myvar', idx).tolist() == data[idx].tolist()

    def test_can_set_attrs_when_creating_main_coord_or_variable(self, mdset):
        mdset.createCoord('mycoord', [1, 2, 3], attrs=dict(myatt=123))
        mdset.createVariable('myvar', [4, 5, 6], 'mycoord', attrs=dict(myatt=9))
        assert mdset.getAttrs('mycoord') == dict(myatt=123)
        assert mdset.getAttrs('myvar') == dict(myatt=9)

    def test_can_add_attrs_after_creating_main_coord_or_variable(self, mdset):
        mdset.createCoord('mycoord', [1, 2, 3], attrs=dict(myatt=123))
        mdset.createVariable('myvar', [4, 5, 6], 'mycoord', attrs=dict(myatt=9))
        mdset.setAttrs('mycoord', dict(myatt2=234))
        mdset.setAttrs('myvar', dict(myatt2=8))
        assert mdset.getAttrs('mycoord') == dict(myatt=123, myatt2=234)
        assert mdset.getAttrs('myvar') == dict(myatt=9, myatt2=8)

    def test_can_set_data_of_main_variable_after_creation(self, mdset):
        mdset.createCoord('mycoord', [1, 2, 3])
        mdset.createVariable('myvar', [4, 5, 6], 'mycoord')
        mdset.setData('myvar', [7, 8, 9])
        assert mdset.getData('myvar').tolist() == [7, 8, 9]

    def test_can_set_partial_data_of_main_variable_after_creation(self, mdset):
        mdset.createCoord('mycoord', [1, 2, 3])
        mdset.createVariable('myvar', [4, 5, 6], 'mycoord')
        mdset.setData('myvar', 7, idx=0)
        assert mdset.getData('myvar').tolist() == [7, 5, 6]
        mdset.setData('myvar', [0, 9], idx=np.s_[1:])
        assert mdset.getData('myvar').tolist() == [7, 0, 9]

    def test_can_set_initial_constant_data_of_multifile_variable_at_creation(self, mdset):
        mdset.createCoord('A', [1, 2, 3])
        mdset.createCoord('B', [4, 5], cross_dataset=True)

        mdset.createVariable('x', 1, ('B', 'A'))
        assert mdset.getData('x').tolist() == [[1, 1, 1], [1, 1, 1]]

    def test_can_set_attrs_of_multifile_variable(self, mdset):
        mdset.createCoord('A', [1, 2, 3])
        mdset.createCoord('B', [4, 5], cross_dataset=True)
        mdset.createVariable('x', 1, ('B', 'A'), attrs=dict(myatt=123))
        assert mdset.getAttrs('x') == dict(myatt=123)

    def test_can_set_initial_slicewise_data_of_multifile_variable_at_creation(self, mdset):
        mdset.createCoord('A', [1, 2, 3])
        mdset.createCoord('B', [4, 5], cross_dataset=True)

        mdset.createVariable('x', np.arange(3).reshape((1, 3)), ('B', 'A'))
        assert mdset.getData('x').tolist() == [[0, 1, 2], [0, 1, 2]]

    def test_can_get_partial_data_of_multifile_variable(self, mdset):
        mdset.createCoord('A', [1, 2, 3, 4])
        mdset.createCoord('B', [5, 6, 7, 8, 9], cross_dataset=True)

        mdset.createVariable('x', np.arange(4), ('B', 'A'))
        assert mdset.getData('x', np.s_[1:3, 1:4]).tolist() == [[1, 2, 3], [1, 2, 3]]

    def test_can_set_partial_data_of_multifile_variable_after_creation(self, mdset):
        mdset.createCoord('A', [1, 2, 3])
        mdset.createCoord('B', [5, 6, 7, 8], cross_dataset=True)

        mdset.createVariable('x', 1, ('B', 'A'))
        mdset.setData('x', [[8], [9]], np.s_[1:3, 1:2])
        data = mdset.getData('x').tolist()
        assert data == [[1, 1, 1], [1, 8, 1], [1, 9, 1], [1, 1, 1]]

    def test_can_set_partial_data_when_multiple_crosscoords(self, mdset):
        mdset.createCoord('A', [1, 2])
        mdset.createCoord('B', [3, 4, 5], cross_dataset=True)
        mdset.createCoord('C', [6, 7, 8, 9])
        mdset.createCoord('D', [8, 7, 6, 5, 4], cross_dataset=True)

        mdset.createVariable('x', 1, ('A', ))
        mdset.createVariable('y', 2, ('B', 'A', ))
        mdset.createVariable('z', 3, ('C', 'B', 'A', ))
        mdset.createVariable('w', 4, ('D', 'C', 'B', 'A', ))

        in_data = np.arange(6).reshape((1, 2, 3, 1))
        idx = np.s_[1::5, :2, -3:, 0:1]
        mdset.setData('w', in_data, idx)
        out_data = mdset.getData('w')
        assert out_data[idx].ravel().tolist() == in_data.ravel().tolist()

    def test_cannot_create_variables_after_subdataset_creation(self, mdset):
        # Create main-dataset coordinate (does not lock dataset)
        mdset.createCoord('A', [1, 2])
        # Create main-dataset variable (does not lock datset)
        mdset.createVariable('x', 1, 'A')
        # Set main-dataset variable data (does not lock dataset)
        mdset.setData('x', 2)
        # Create cross-dataset coordinate (does not lock dataset)
        mdset.createCoord('B', [1, 2, 3], cross_dataset=True)
        # Create cross-dataset variable (does not lock dataset)
        mdset.createVariable('y', 3, 'B')
        # Access cross-dataset variable (DOES lock dataset)
        mdset.setData('y', 4)
        # Cannot create new main-dataset coordinates when locked
        with pytest.raises(TypeError):
            mdset.createCoord('C', [1, 2])
        # Cannot create new cross-dataset coordinates when locked
        with pytest.raises(TypeError):
            mdset.createCoord('D', [1, 2], cross_dataset=True)
        # Cannot create new main-dataset variables when locked
        with pytest.raises(TypeError):
            mdset.createVariable('z', 5, 'A')
        # Cannot create new cross-dataset variables when locked
        with pytest.raises(TypeError):
            mdset.createVariable('w', 6, 'B')
        # CAN access variable data after locking
        mdset.setData('x', 7)
        assert mdset.getData('x').tolist() == [7, 7]
        mdset.setData('y', 8)
        assert mdset.getData('y').tolist() == [8, 8, 8]

    def test_cannot_set_crossvar_attributes_after_subdataset_creation(self, mdset):
        # Create main-dataset coordinate (does not lock dataset)
        mdset.createCoord('main_coord', [1, 2])
        # Set main-dataset coordinate attributes (does not lock dataset)
        mdset.setAttrs('main_coord', dict(main_coord_att=2))
        # Create main-dataset variable (does not lock datset)
        mdset.createVariable('main_var', 1, 'main_coord')
        # Set main-dataset variable attributes (does not lock dataset)
        mdset.setAttrs('main_var', dict(main_var_att=3))
        # Create cross-dataset coordinate (does not lock dataset)
        mdset.createCoord('cross_coord', [1, 2, 3], cross_dataset=True)
        # Set cross-dataset coordinate attributes (does not lock dataset)
        mdset.setAttrs('cross_coord', dict(cross_coord_att=4))
        # Create cross-dataset variable (does not lock dataset)
        mdset.createVariable('cross_var', 3, 'cross_coord')
        # Set cross-dataset variable attributes (does not lock dataset)
        mdset.setAttrs('cross_var', dict(cross_var_att=5))
        # Access cross-dataset variable (DOES lock dataset)
        mdset.setData('cross_var', 4)
        # Cannot edit any attributes when locked
        with pytest.raises(TypeError):
            mdset.setAttrs('main_coord', dict(newatt=3))
        with pytest.raises(TypeError):
            mdset.setAttrs('main_var', dict(newatt=3))
        with pytest.raises(TypeError):
            mdset.setAttrs('cross_var', dict(newatt=3))
        with pytest.raises(TypeError):
            mdset.setAttrs('cross_coord', dict(newatt=3))

    def test_accepts_in_memory_datasets(self):
        import netCDF4 as nc
        with nc.Dataset(uuid4(), mode='w', diskless=True) as dset:
            mdset = output.MultiDataset(dset)

            mdset.createCoord('main_coord', [1, 2])
            assert dset.variables['main_coord'][:].tolist() == [1, 2]

            mdset.createVariable('main_var', 1, 'main_coord')
            assert dset.variables['main_var'][:].tolist() == [1, 1]

            mdset.setData('main_var', 2, idx=[0])
            assert dset.variables['main_var'][:].tolist() == [2, 1]

            # Can read after calling close method
            # (Since underlying dataset is not closed)
            mdset.close()
            assert dset.variables['main_var'][:].tolist() == [2, 1]


class Test_nc_to_dict:
    def test_returns_valid_xarray_dict_representation(self):
        with netCDF4.Dataset(uuid4(), 'w', diskless=True) as dset:
            dset.createDimension('mydim1', 2)
            dset.createDimension('mydim2', 3)
            dset.createVariable('myvar1', int, 'mydim1')[:] = 0
            dset.createVariable('myvar2', int, ('mydim1', 'mydim2'))[:] = 1
            dset['myvar1'].long_name = "Variable 2"
            d = output.nc_to_dict(dset)

        xr_dset = xr.Dataset.from_dict(d)
        assert list(xr_dset.variables) == ['myvar1', 'myvar2']
        assert xr_dset.myvar2.values.tolist() == [[1, 1, 1], [1, 1, 1]]
        assert xr_dset.myvar1.attrs['long_name'] == 'Variable 2'



import contextlib
import datetime
import numpy as np
import pytest
from ladim_aggregate import histogram
import xarray as xr


class Test_Histogrammer:
    def test_can_generate_histogram_piece_from_chunk(self):
        h = histogram.Histogrammer(
            bins=dict(
                z=dict(edges=[-1.5, 1.5, 4.5], centers=[0, 3]),
                y=dict(edges=[-1, 1, 3, 5], centers=[0, 2, 4]),
                x=dict(edges=[-.5, .5, 1.5, 2.5, 3.5], centers=[0, 1, 2, 3]),
            )
        )
        chunk = xr.Dataset(dict(x=[0, 1, 3], y=[0, 2, 4], z=[0, 1, 3]))
        hist_piece = next(h.make(chunk))
        assert hist_piece['values'].tolist() == [
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0]],
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1]],
        ]
        start = [idx.start for idx in hist_piece['indices']]
        assert start == [0, 0, 0]
        stop = [idx.stop for idx in hist_piece['indices']]
        assert stop == [2, 3, 4]

    def test_can_generate_weighted_histogram_piece_from_chunk(self):
        h = histogram.Histogrammer(bins=dict(x=dict(
            edges=[0, 2, 6], centers=[1, 4],
        )))
        chunk = xr.Dataset(dict(x=[1, 3, 5], _auto_weights=[10, 100, 1000]))
        hist_piece = next(h.make(chunk))
        assert hist_piece['values'].tolist() == [10, 1100]


class Test_adaptive_histogram:
    def test_returns_same_as_histogramdd_if_count(self):
        sample = [
            [1.5, 2.5, 3.5, 4.5, 5.5],
            [6.5, 6.5, 7.5, 8.5, 9.5],
        ]
        bins = [[1, 4, 6], [6, 8, 9, 10, 12]]

        hist, _ = np.histogramdd(sample, bins)

        hist2 = np.zeros([len(b) - 1 for b in bins])
        hist_chunk, idx = histogram.adaptive_histogram(sample, bins)
        assert idx == np.s_[0:2, 0:3]
        hist2[idx] = hist_chunk

        assert hist2.tolist() == hist.tolist()

    def test_returns_only_partial_matrix(self):
        sample = [
            [1.5, 2.5, 3.5, 4.5, 5.5],
            [6.5, 6.5, 7.5, 8.5, 9.5],
        ]
        bins = [[1, 4, 6, 8], [6, 8, 9, 10, 12]]

        hist_chunk, idx = histogram.adaptive_histogram(sample, bins)

        assert hist_chunk.shape[0] < len(bins[0]) - 1
        assert hist_chunk.shape[1] < len(bins[1]) - 1

    def test_returns_same_as_histogramdd_if_weights(self):
        sample = [
            [1.5, 2.5, 3.5, 4.5, 5.5],
            [6.5, 6.5, 7.5, 8.5, 9.5],
        ]
        weights = [1, 2, 3, 4, 5]
        bins = [[1, 4, 6], [6, 8, 9, 10, 12]]

        hist, _ = np.histogramdd(sample, bins, weights=weights)

        hist2 = np.zeros([len(b) - 1 for b in bins])
        hist_chunk, idx = histogram.adaptive_histogram(sample, bins, weights=weights)
        assert idx == np.s_[0:2, 0:3]
        hist2[idx] = hist_chunk

        assert hist2.tolist() == hist.tolist()

    def test_returns_same_as_histogramdd_if_no_particles(self):
        sample = [[], []]
        bins = [[1, 4, 6], [6, 8, 9, 10, 12]]

        hist, _ = np.histogramdd(sample, bins)

        hist2 = np.zeros([len(b) - 1 for b in bins])
        hist_chunk, idx = histogram.adaptive_histogram(sample, bins)
        hist2[idx] = hist_chunk

        assert hist2.tolist() == hist.tolist()

    def test_returns_same_as_histogramdd_if_particles_outside_range(self):
        # Weights are included since a previous bug caused the function to
        # fail when weighted particles were outside range
        sample = [[1, 2, 3, 3, 4], [5, 6, 7, 8, 9]]
        weights = [1, 1, 1, 1, 1]
        bins = [[1.5, 2.5, 3.5], [6.5, 7.5, 8.5, 9.5]]
        hist_np, _ = np.histogramdd(sample, bins, weights=weights)

        hist2 = np.zeros([len(b) - 1 for b in bins])
        hist_chunk, idx = histogram.adaptive_histogram(sample, bins, weights=weights)
        assert idx == np.s_[1:2, 0:2]
        hist2[idx] = hist_chunk

        assert hist2.tolist() == hist_np.tolist()

    def test_returns_same_as_histogramdd_if_no_particles_in_range(self):
        sample = [[1, 2, 3, 3, 4], [5, 6, 7, 8, 9]]
        bins = [[10, 20], [30, 40, 50]]
        hist_np, _ = np.histogramdd(sample, bins)

        hist2 = np.zeros([len(b) - 1 for b in bins])
        hist_chunk, idx = histogram.adaptive_histogram(sample, bins)
        assert idx == np.s_[1:0, 1:0]
        hist2[idx] = hist_chunk

        assert hist2.tolist() == hist_np.tolist()


class Test_autobins:
    def test_computes_centers_if_spec_is_list(self):
        spec = dict(x=[1, 2, 3])
        bins = histogram.autobins(spec, dset=None)
        assert bins['x']['edges'].tolist() == [1, 2, 3]
        assert bins['x']['centers'].tolist() == [1.5, 2.5]

    def test_returns_verbatim_if_spec_is_edges_labels(self):
        spec = dict(x=dict(edges=[1, 2, 3], labels=[10, 20]))
        bins = histogram.autobins(spec, dset=None)
        assert bins['x']['edges'].tolist() == [1, 2, 3]
        assert bins['x']['centers'].tolist() == [10, 20]

    def test_returns_inclusive_range_if_spec_is_min_max_step(self):
        spec = dict(x=dict(min=1, max=10, step=3))
        bins = histogram.autobins(spec, dset=None)
        assert bins['x']['edges'].tolist() == [1, 4, 7, 10]
        assert bins['x']['centers'].tolist() == [2.5, 5.5, 8.5]

    def test_accepts_multiple_specs(self):
        spec_1 = dict(x=dict(min=1, max=10, step=3))
        bins_1 = histogram.autobins(spec_1, dset=None)
        spec_2 = dict(y=[1, 2, 3])
        bins_2 = histogram.autobins(spec_2, dset=None)
        spec = {**spec_1, **spec_2}
        bins = histogram.autobins(spec, dset=None)
        assert bins['x']['edges'].tolist() == bins_1['x']['edges'].tolist()
        assert bins['y']['edges'].tolist() == bins_2['y']['edges'].tolist()

    def test_returns_aligned_range_if_resolution(self):
        class MockLadimDataset:
            def __init__(self):
                self._specials = {'MIN_x': 10, 'MAX_x': 19}
                self.add_aggregation_variable = lambda v, op: f'{op.upper()}_{v}'

            def get_aggregation_value(self, key):
                return self._specials[key]

        spec = dict(x=3)
        bins = histogram.autobins(spec, dset=MockLadimDataset())
        assert bins['x']['edges'].tolist() == [9, 12, 15, 18, 21]

    def test_returns_bins_if_unique(self):
        class MockLadimDataset:
            def __init__(self):
                self._specials = {'UNIQUE_x': [1, 2, 5]}
                self.add_aggregation_variable = lambda v, op: f'{op.upper()}_{v}'

            def get_aggregation_value(self, key):
                return self._specials[key]

        spec = dict(x='unique')
        bins = histogram.autobins(spec, dset=MockLadimDataset())
        assert bins['x']['edges'].tolist() == [1, 2, 5, 6]
        assert bins['x']['centers'].tolist() == [1, 2, 5]

    def test_copies_attributes_from_input_dataset(self):
        class MockLadimDataset:
            def __init__(self):
                self.attributes = dict(
                    x=dict(long_name="x coordinate value"),
                    y=dict(long_name="y coordinate value"),  # An extra attribute which is not in the bins
                )

        spec = dict(x=[1, 2, 3])
        bins = histogram.autobins(spec, dset=MockLadimDataset())
        assert bins['x']['attrs']['long_name'] == "x coordinate value"
        assert 'y' not in bins

    def test_converts_time_specs(self):
        class MockLadimDataset:
            @contextlib.contextmanager
            def open_dataset(self, _):
                attrs = dict(units='hours since 1980-01-01')
                tvar = xr.Variable(dims='time', data=[], attrs=attrs)
                yield xr.Dataset(data_vars=dict(time=tvar))

        spec = dict(time=dict(min='1980-01-01', max='1980-01-03', step='1 days'))
        bins = histogram.autobins(spec, dset=MockLadimDataset())
        assert bins['time']['centers'].tolist() == [12, 36]
        assert bins['time']['edges'].tolist() == [0, 24, 48]


class Test_convert_datebins:
    @pytest.fixture(scope='class')
    def time_dset(self):
        attrs = dict(units='days since 2000-01-02')
        tvar = xr.Variable(dims='time', data=[1, 3, 5], attrs=attrs)
        xvar = xr.Variable(dims='time', data=[10, 30, 50])
        return xr.Dataset(data_vars=dict(time=tvar, X=xvar))

    def test_converts_only_datevars(self, time_dset):
        spec = dict(
            time=['2000-01', '2000-02', '2000-03'],
            X=[0, 10, 20, 30],
        )
        newspec = histogram.convert_datebins(spec, time_dset)
        assert newspec == dict(
            time=[-1, 30, 59],
            X=[0, 10, 20, 30],
        )

    def test_ignores_nonexisting_vars(self, time_dset):
        spec = dict(Y=[0, 10, 20, 30])
        newspec = histogram.convert_datebins(spec, time_dset)
        assert newspec == spec


class Test_convert_binspec:
    def test_converts_edgeformat(self):
        spec = ['1970-01', '1970-02', '1970-03', 100]
        result = histogram.convert_binspec(spec, 'days since 1970-01-01', 'standard')
        assert result == [0, 31, 31+28, 100]

    def test_converts_rangeformat(self):
        spec = dict(min='1970-01', max='1970-02', step='48 hours')
        result = histogram.convert_binspec(spec, 'days since 1970-01-01', 'standard')
        assert result == dict(min=0, max=31, step=2)

    def test_converts_labelformat(self):
        spec = dict(edges=['1970-01', '1970-02', '1970-03'], labels=['jan', 'feb'])
        result = histogram.convert_binspec(spec, 'days since 1970-01-01', 'standard')
        assert result == dict(edges=[0, 31, 31+28], labels=['jan', 'feb'])

    def test_converts_cfstring_format(self):
        spec = '48 hours'
        result = histogram.convert_binspec(spec, 'days since 2020-01-01', 'standard')
        assert result == 2


class Test_convert_step:
    def test_returns_integers_verbatim(self):
        result = histogram.convert_step(123, 'days since 1980-01-01', 'standard')
        assert result == 123

    def test_converts_string_spec(self):
        result = histogram.convert_step('48 hours', 'days since 1980-01-01', 'standard')
        assert result == 2

    def test_converts_string_spec_when_timedate(self):
        result = histogram.convert_step('1 days', 'hours since 1980-01-01 01:00:00', 'standard')
        assert result == 24


class Test_convert_date:
    def test_returns_integers_verbatim(self):
        result = histogram.convert_date(123, 'days since 1970-01-01', 'standard')
        assert result == 123

    def test_returns_floats_verbatim(self):
        result = histogram.convert_date(123.5, 'days since 1970-01-01', 'standard')
        assert result == 123.5

    def test_converts_dates(self):
        date = datetime.date(2000, 5, 3)
        result = histogram.convert_date(date, 'days since 1970-01-01', 'standard')
        assert result == 11080

    def test_converts_datetimes(self):
        date = datetime.datetime(2000, 5, 3, 12, 0, 0)
        result = histogram.convert_date(date, 'days since 1970-01-01', 'standard')
        assert result == 11080

    def test_converts_datetime64(self):
        date = np.datetime64('1970-02')
        result = histogram.convert_date(date, 'days since 1970-01-01', 'standard')
        assert result == 31

    def test_converts_string(self):
        date = '1970-02'
        result = histogram.convert_date(date, 'days since 1970-01-01', 'standard')
        assert result == 31


class Test_align_to_resolution:
    def test_aligns_to_integer(self):
        assert histogram.align_to_resolution(value=22, resolution=3) == 21
        assert histogram.align_to_resolution(value=21, resolution=3) == 21
        assert histogram.align_to_resolution(value=19, resolution=3) == 18

    def test_aligns_to_time(self):
        align = histogram.align_to_resolution

        date = np.datetime64('2020-01-02T03:04:05')

        second = np.timedelta64(1, 's')
        minute = np.timedelta64(1, 'm')
        hour = np.timedelta64(1, 'h')
        day = np.timedelta64(1, 'D')

        assert align(date, second).astype(str) == '2020-01-02T03:04:05.000000'
        assert align(date, minute).astype(str) == '2020-01-02T03:04:00.000000'
        assert align(date, hour).astype(str) == '2020-01-02T03:00:00.000000'
        assert align(date, day).astype(str) == '2020-01-02T00:00:00.000000'


class Test_t64conv:
    def test_returns_timedeltas_verbatim(self):
        t = np.timedelta64(1, 's')
        result = histogram.t64conv(t)
        assert result is t

    def test_returns_numbers_verbatim(self):
        t = 123
        result = histogram.t64conv(t)
        assert result is t

    def test_converts_tuple_values(self):
        t = [23, 's']
        result = histogram.t64conv(t)
        assert result == np.timedelta64(23, 's')

    def test_converts_cftime_strings(self):
        t_vec = [
            '1 day',
            '24 hours',
            '24 h',
            '20 minutes',
            '23 seconds',
            '23 s',
            '23000 ms',
            '23000000 us',
        ]
        result = [histogram.t64conv(t) for t in t_vec]
        assert [r.astype(str) for r in result] == [
            '1 days',
            '24 hours',
            '24 hours',
            '20 minutes',
            '23 seconds',
            '23 seconds',
            '23000 milliseconds',
            '23000000 microseconds',
        ]


class Object:
    pass



from ladim_aggregate import geotag
import json
import xarray as xr
import pkgutil


class Test_create_geotagger:
    def test_can_return_correct_polygon_attribute_of_particle(self):
        chunk = xr.Dataset(
            data_vars=dict(
                lon=xr.Variable('pid', [.5, .5, 10.5]),
                lat=xr.Variable('pid', [60.5, 70.5, 70.5]),
            )
        )

        pkg = 'ladim_aggregate.examples.connect'
        geojson = json.loads(pkgutil.get_data(pkg, 'regions.geojson').decode('utf-8'))

        geotagger = geotag.create_geotagger(
            attribute="region",
            x_var='lon',
            y_var='lat',
            geojson=geojson,
            missing=-1,
        )

        region = geotagger(chunk)
        assert region.dims == ('pid', )
        assert region.values.tolist() == [101, -1, 102]



from ladim_aggregate import examples
import netCDF4 as nc
from uuid import uuid4
import pytest


class Test_Example:
    def test_available(self):
        available = examples.Example.available()
        assert 'grid_2D' in available

    def test_config(self):
        ex = examples.Example('grid_2D')
        assert 'bins' in ex.config

    def test_package(self):
        ex = examples.Example('grid_2D')
        assert ex.package == 'ladim_aggregate.examples.grid_2D'

    def test_descr(self):
        ex = examples.Example('grid_2D')
        assert ex.descr.startswith('Basic example')

    def test_files(self):
        ex = examples.Example('grid_2D')
        data_dict = ex.files()
        assert hasattr(data_dict['aggregate.yaml'], 'decode')  # byte object
        assert hasattr(data_dict['ladim.nc.yaml'], 'decode')  # byte object
        assert hasattr(data_dict['ladim.nc'], 'to_netcdf')  # xarray object

    def test_extract(self, tmp_path):
        ex = examples.Example('grid_2D')
        ex.extract(tmp_path)
        fnames = {f.name for f in tmp_path.glob('*')}
        assert 'ladim.nc' in fnames
        assert 'aggregate.yaml' in fnames
        assert 'count.nc' not in fnames       # Output files not included
        assert 'ladim.nc.yaml' not in fnames  # Only unpacked files are included


class Test_nc_dump:
    def test_returns_dict_when_netcdf_input(self):
        with nc.Dataset(uuid4(), 'w', diskless=True) as dset:
            dset = dset  # type: nc.Dataset
            dset.createDimension('time', 3)

            v = dset.createVariable('count', int, 'time')
            v[:] = [4, 2, 2]
            v.long_name = 'Count'

            netcdf_as_dict = examples.nc_dump(dset)

        assert netcdf_as_dict == dict(
            count=dict(dims='time', data=[4, 2, 2], attrs=dict(long_name='Count')),
        )


named_examples = examples.Example.available()


class Test_run:
    @pytest.mark.parametrize("name", named_examples)
    def test_matches_output(self, name):
        ex = examples.Example(name)
        result, expected = ex.run()
        assert result == expected



import contextlib
import logging
import glob
import numpy as np
import xarray as xr
import typing
import re


logger = logging.getLogger(__name__)


class LadimInputStream:
    def __init__(self, spec):
        self.datasets = glob_files(spec)
        self._attributes = None
        self._derived_variables = dict()
        self._agg_variables = dict()
        self._init_variables = []

    @property
    def attributes(self):
        if self._attributes is None:
            self._attributes = dict()
            with self.open_dataset(0) as dset:
                logger.debug('Read attributes')
                for k, v in dset.variables.items():
                    self._attributes[k] = v.attrs
        return self._attributes

    def open_dataset(self, idx: int) -> xr.Dataset:
        """
        Open ladim dataset by index.

        :param idx: Ordinal index of the dataset
        :return: The dataset
        """
        return _open_spec(self.datasets[idx])

    def add_derived_variable(self, varname, definition):
        """
        Define new variables from expressions based on the old variables
        :param varname: Variable name
        :param definition: Variable definition
        """
        self._derived_variables[varname] = create_varfunc(definition)

    def add_aggregation_variable(self, varname, operator):
        """
        Aggregation variables are the result of aggregation operations such as `max` or
        `min`. They require a scanning of the whole dataset. The scanning is deferred
        until a specific aggregate value is requested.

        Variables defined here are available through the function get_aggregation_value

        :param varname: Name of the variable
        :param operator: Name of the operator
        :return: The name of the aggregate variable (e.g., MAX_temp)
        """
        key = operator.upper() + '_' + varname
        self._agg_variables[key] = dict(
            key=key,
            varname=varname,
            operator=operator,
            value=None,
        )
        return key

    def get_aggregation_value(self, key):
        if self._agg_variables[key]['value'] is None:
            self._update_agg_variables()
        return self._agg_variables[key]['value']

    def add_init_variable(self, varname):
        """
        Init variables are derived variables that do not need a pre-scanning of the
        dataset. The value of an init variable is the first value found when looping
        through the time steps.

        Variables added here are available through the chunks() function as variables
        named like `<varname>_INIT`.

        :param varname: Variable name
        :return: None
        """
        self._init_variables.append(varname)

    def add_grid_variable(self, data_array, method):
        """
        Grid variables are derived variables that are interpolated from a grid. They
        are added to the dataset through the chunks() function.

        :param data_array: A named xarray DataArray defining the variable values on a grid
        :param method: Either 'linear', 'nearest' or 'bin'
        :return: None
        """
        self._derived_variables[data_array.name] = create_varfunc(('grid', data_array, method))

    def _update_agg_variables(self):
        # Find all unassigned aggfuncs and store them variable-wise
        spec_keys = [k for k, v in self._agg_variables.items() if v['value'] is None]
        spec = {}
        for k in spec_keys:
            opname = self._agg_variables[k]['operator']
            vname = self._agg_variables[k]['varname']
            spec[vname] = spec.get(vname, []) + [opname]

        if spec == {}:
            return

        out = self.scan(spec)

        for k in spec_keys:
            opname = self._agg_variables[k]['operator']
            vname = self._agg_variables[k]['varname']
            value = out[vname][opname]
            if opname in ['max', 'min']:
                xr_var = xr.Variable((), value)
            elif opname == 'unique':
                xr_var = xr.Variable(k, value)
            else:
                raise NotImplementedError
            self._agg_variables[k]['value'] = xr_var

    def scan(self, spec):
        """
        Scan the dataset and return summary values for chosen variables.

        The input parameter `spec` is used to specify summary functions. Each variable
        name can be mapped to one or more functions.

        :param spec: A mapping of variable names to summary function names.
        :returns: A mapping of variable names to summary function name/value pairs.
        """
        out = {k: {fun: None for fun in funclist} for k, funclist in spec.items()}

        def agg_log(aggfunc, aggval):
            if aggfunc == "unique":
                logger.debug(f'Number of unique values: {len(aggval)}')
            elif aggfunc == "max":
                logger.debug(f'Max value: {aggval}')
            elif aggfunc == "min":
                logger.debug(f'Min value: {aggval}')

        def update_output(ddset, sub_spec):
            for varname, funclist in sub_spec.items():
                if varname in ddset.variables:
                    logger.debug(f'Load "{varname}" values')
                    data = ddset.variables[varname].values
                elif varname in self._derived_variables:
                    logger.debug(f'Compute "{varname}" values')
                    fn = self._derived_variables[varname]
                    data = fn(ddset).values
                else:
                    raise ValueError(f'Unknown variable name: "{varname}"')

                for fun in funclist:
                    out[varname][fun] = update_agg(out[varname][fun], fun, data)
                    agg_log(fun, out[varname][fun])

        # Particle variables do only need the first dataset
        with self.open_dataset(0) as dset:
            update_output(dset, spec)
            pvars = [k for k in spec if k in dset and dset[k].dims == ('particle', )]
            spec_without_particle_vars = {k: v for k, v in spec.items() if k not in pvars}

        if spec_without_particle_vars:
            for idx in range(1, len(self.datasets)):
                with self.open_dataset(idx) as dset:
                    update_output(dset, spec_without_particle_vars)

        return out

    def chunks(self, filters=None, timestep_filter=None, particle_filter=None) -> typing.Iterator[xr.Dataset]:
        """
        Return one ladim timestep at a time.

        Variables associated with "time" or "particle" are distributed correctly over
        the particles present in the current timestep.

        For each time step, an optional filter is applied. Furthermore, the derived
        variables added to the dataset are computed.

        :param filters: A filtering expression
        :param timestep_filter: A set of time steps indices to return. If None (default),
        return all timesteps. The time steps list is sorted before being applied, and
        nonunique elements are disregarded.
        :param particle_filter: The same as `filter`, except that it only returns a
        single instance per particle (i.e., the first occurrence when the condition is
        met).
        :return: An xarray dataset indexed by "pid" for each time step.
        """
        filterfn = create_varfunc(filters)

        if particle_filter is not None:
            particle_filterfn = create_varfunc(('pfilter', particle_filter))
        else:
            particle_filterfn = None

        # Initialize the "init variables"
        init_variables = {k: None for k in self._init_variables}

        for chunk in ladim_iterator(self.datasets, timestep_filter):
            # Apply filter
            filter_idx = None
            num_unfiltered = chunk.dims['pid']
            if filterfn:
                logger.debug("Apply filter")
                filter_idx = filterfn(chunk).values
                num_unfiltered = np.count_nonzero(filter_idx)
                logger.debug(f'Number of remaining particles: {num_unfiltered}')

            if (num_unfiltered == 0) and (len(init_variables) == 0):
                continue

            # Add derived variables (such as weights and geotags)
            for varname, fn in self._derived_variables.items():
                logger.debug(f'Compute "{varname}"')
                chunk = chunk.assign(**{varname: fn(chunk)})

            # Add init variables (such as region_INIT)
            for varname, data_and_mask in init_variables.items():
                pid = chunk['pid'].values
                input_data = (chunk[varname].values, pid)
                data, mask = update_init(data_and_mask, input_data)
                init_variables[varname] = (data, mask)
                xr_var = xr.Variable('pid', data[pid])
                chunk = chunk.assign(**{f"{varname}_INIT": xr_var})

            # Add aggregation variables (such as MAX_temp)
            for varname in self._agg_variables:
                xr_var = self.get_aggregation_value(varname)
                chunk = chunk.assign(**{varname: xr_var})

            # Add particle filtering
            if particle_filterfn is not None:
                logger.debug("Apply particle filter")
                pfilter_idx = particle_filterfn(chunk).values
                if filterfn:
                    filter_idx &= pfilter_idx
                else:
                    filter_idx = pfilter_idx
                num_unfiltered = np.count_nonzero(filter_idx)
                logger.debug(f'Number of remaining particles: {num_unfiltered}')

            # Do actual filtering
            if filter_idx is not None:
                chunk = chunk.isel(pid=filter_idx)

            yield chunk


def get_time(timevar):
    return xr.decode_cf(timevar.to_dataset(name='timevar')).timevar.values


def get_varfunc_from_numexpr(spec):
    import numexpr
    ex = numexpr.NumExpr(spec)

    def weight_fn(chunk):
        args = []
        for n in ex.input_names:
            logger.debug(f'Load variable "{n}"')
            args.append(chunk[n].values)
        logger.debug(f'Compute expression "{spec}"')
        return xr.Variable('pid', ex.run(*args))

    return weight_fn


def get_varfunc_from_callable(fn):
    import inspect
    signature = inspect.signature(fn)

    def weight_fn(chunk):
        args = [chunk[n].values for n in signature.parameters.keys()]
        return xr.Variable('pid', fn(*args))

    return weight_fn


def get_varfunc_from_funcstring(s: str):
    import importlib
    module_name, func_name = s.rsplit('.', 1)
    module = importlib.import_module(module_name)
    func = getattr(module, func_name)
    return get_varfunc_from_callable(func)


def get_varfunc_from_grid(darr: xr.DataArray, method):
    def fn_other(chunk):
        coords = {d: chunk.variables[d] for d in darr.dims}
        return darr.interp(coords, method=method).variable

    def fn_bin(chunk):
        idx = {}
        for d in darr.dims:
            dvalues = darr[d].values
            pvalues = chunk.variables[d].values
            data = np.searchsorted(dvalues, pvalues, side='right') - 1
            idx[d] = xr.Variable(dims='pid', data=data)
        return darr.isel(idx).variable

    if method == 'bin':
        return fn_bin
    else:
        return fn_other


def glob_files(spec):
    """
    Convert a set of glob patterns to a list of files

    :param spec: One or more glob patterns
    :return: A list of files
    """

    # Convert input spec to a sequence
    if isinstance(spec, tuple) or isinstance(spec, list):
        specs = spec
    else:
        specs = [spec]

    # Expand glob patterns in spec
    files = []
    for s in specs:
        if isinstance(s, str):
            files += sorted(glob.glob(s))
        else:
            files.append(s)

    return files


def dset_iterator(specs):
    for spec in specs:
        with _open_spec(spec) as dset:
            yield dset


def ladim_iterator(ladim_dsets, timesteps=None):
    """
    Return one chunk of a ladim dataset per timestep

    Distributes particle and time variable over the `particle_instance` dimension.

    :param ladim_dsets: A sequence of ladim datasets
    :param timesteps: A set of time steps indices to return. If None (default), return
    all timesteps. The time steps list is sorted before being applied, and nonunique
    elements are disregarded.
    :return: An iterator over timestep chunks
    """

    # Prepare to filter out some time steps
    tidx_start = 0
    if timesteps is not None:
        timesteps = np.unique(timesteps)

    for dset in dset_iterator(ladim_dsets):
        instance_offset = dset.get('instance_offset', 0)
        pcount_cum = np.concatenate([[0], np.cumsum(dset.particle_count.values)])
        tidx_list = range(dset.dims['time'])

        # Potentially filter out some time steps
        if timesteps is not None:
            include_steps = timesteps - tidx_start
            tidx_list = np.intersect1d(tidx_list, include_steps)
            tidx_start += dset.dims['time']

        for tidx in tidx_list:
            timestr = str(get_time(dset.time[tidx]).astype('datetime64[s]')).replace("T", " ")
            logger.info(f'Read time step {timestr} (time={dset.time[tidx].values.item()})')
            iidx = slice(pcount_cum[tidx], pcount_cum[tidx + 1])
            logger.debug(f'Number of particles: {iidx.stop - iidx.start}')
            if iidx.stop == iidx.start:
                continue

            pid = xr.Variable('pid', dset.pid[iidx].values, dset.pid.attrs)

            ddset = xr.Dataset(
                data_vars=dict(instance_offset=instance_offset + iidx.start),
                coords=dict({pid.dims[0]: pid}),
                attrs=dset.attrs,
            )

            for k, v in dset.variables.items():
                if k in ('pid', 'instance_offset'):
                    continue

                logger.debug(f'Load variable "{k}"')
                if v.dims == ('particle_instance', ):
                    new_var = xr.Variable(pid.dims[0], v[iidx].values, v.attrs)
                elif v.dims == ('particle', ):
                    new_var = xr.Variable(pid.dims[0], v.values[pid.values], v.attrs)
                elif v.dims == ('time', ):
                    data = np.broadcast_to(v.values[tidx], (iidx.stop - iidx.start, ))
                    new_var = xr.Variable(pid.dims[0], data, v.attrs)
                else:
                    raise ValueError(f'Unknkown dimension: "{v.dims}"')

                ddset = ddset.assign(**{k: new_var})

            yield ddset


def update_agg(old, aggfun, data):
    funcs = dict(
        max=update_max, min=update_min, unique=update_unique, init=update_init,
        final=update_final,
    )
    return funcs[aggfun](old, data)


def update_init(old, data, final=False):
    if old is None:
        old = (np.zeros(0, dtype=data[0].dtype), np.zeros(0, dtype=bool))

    # Unpack input arguments
    old_data, old_mask = old
    new_data, new_pid = data

    # Expand array if necessary
    max_pid = np.max(new_pid)
    if max_pid >= len(old_data):
        old_data2 = np.zeros(max_pid + 1, dtype=old_data.dtype)
        old_data2[:len(old_data)] = old_data
        old_mask2 = np.zeros(max_pid + 1, dtype=bool)
        old_mask2[:len(old_mask)] = old_mask
        old_data = old_data2
        old_mask = old_mask2

    if not final:
        # Filter out the new particles
        # Flip because the least recent pid number should be used
        is_unexisting_pid = ~old_mask[new_pid]
        new_data = np.flip(new_data[is_unexisting_pid])
        new_pid = np.flip(new_pid[is_unexisting_pid])

    # Store the new particles
    old_data[new_pid] = new_data
    old_mask[new_pid] = True

    return old_data, old_mask


def update_final(old, data):
    return update_init(old, data, final=True)


def update_max(old, data):
    if old is None:
        return np.max(data)
    else:
        return max(np.max(data), old)


def update_min(old, data):
    if old is None:
        return np.min(data)
    else:
        return min(np.min(data), old)


def update_unique(old, data):
    if old is None:
        return np.unique(data).tolist()
    else:
        unq = np.unique(data)
        return np.union1d(old, unq).tolist()


@contextlib.contextmanager
def _open_spec(spec):
    if isinstance(spec, str):
        logger.debug(f'Open dataset "{spec}"')
        with xr.open_dataset(spec, decode_cf=False) as ddset:
            yield ddset
            logger.debug(f'Close dataset "{spec}"')
    else:
        logger.debug(f'Enter new dataset')
        yield spec


def create_varfunc(spec):
    if spec is None:
        return None
    elif isinstance(spec, tuple) and spec[0] == 'geotag':
        from .geotag import create_geotagger
        return create_geotagger(**spec[1])
    elif isinstance(spec, tuple) and spec[0] == 'pfilter':
        return create_pfilter(spec[1])
    elif isinstance(spec, tuple) and spec[0] == 'grid':
        return get_varfunc_from_grid(spec[1], spec[2])
    elif isinstance(spec, str):
        if re.match(r'[.\w]+\.\w+$', spec):
            return get_varfunc_from_funcstring(spec)
        else:
            return get_varfunc_from_numexpr(spec)
    elif callable(spec):
        return get_varfunc_from_callable(spec)
    else:
        raise TypeError(f'Unknown type: {type(spec)}')


def create_pfilter(spec):
    fn = create_varfunc(spec)
    has_been_triggered = ResizableArray(100, dtype=bool)

    def pfilter(chunk):
        fn_val = fn(chunk)
        pid = chunk['pid'].values
        max_pid = pid.max()
        if max_pid > len(has_been_triggered):
            has_been_triggered.resize(max_pid + 1)
        is_new = ~has_been_triggered.data[pid]
        condition = fn_val & is_new
        pid_triggered = pid[condition]
        has_been_triggered.data[pid_triggered] = True
        return xr.Variable('pid', condition)

    return pfilter


class ResizableArray:
    def __init__(self, capacity, dtype):
        self.data = np.zeros(capacity, dtype=dtype)

    def resize(self, new_capacity):
        old_data = self.data
        self.data = np.zeros(new_capacity, dtype=old_data.dtype)
        self.data[:len(old_data)] = old_data

    def __len__(self):
        return len(self.data)



from ladim_aggregate import input as ladim_input
import numpy as np
import pytest
import xarray as xr
from unittest.mock import patch


@pytest.fixture(scope='module')
def ladim_dset():
    return xr.Dataset(
        data_vars=dict(
            X=xr.Variable('particle_instance', [5, 5, 6, 6, 5, 6]),
            Y=xr.Variable('particle_instance', [60, 60, 60, 61, 60, 62]),
            Z=xr.Variable('particle_instance', [0, 1, 2, 3, 4, 5],
                          attrs=dict(standard_name='depth')),
            lon=xr.Variable('particle_instance', [5, 5, 6, 6, 5, 6]),
            lat=xr.Variable('particle_instance', [60, 60, 60, 61, 60, 62]),
            instance_offset=xr.Variable((), 0),
            farm_id=xr.Variable('particle', [12345, 12346, 12347, 12348]),
            pid=xr.Variable('particle_instance', [0, 1, 2, 3, 1, 2]),
            particle_count=xr.Variable('time', [4, 2]),
        ),
        coords=dict(
            time=np.array(['2000-01-02', '2000-01-03']).astype('datetime64[D]'),
        ),
    )


@pytest.fixture(scope='class')
def ladim_dset2(ladim_dset):
    d = ladim_dset.copy(deep=True)
    d['instance_offset'] += d.dims['particle_instance']
    d = d.assign_coords(time=d.time + np.timedelta64(2, 'D'))
    return d


class Test_ladim_iterator:
    def test_returns_one_dataset_per_timestep_when_multiple_datasets(self, ladim_dset, ladim_dset2):
        it = ladim_input.ladim_iterator([ladim_dset, ladim_dset2])
        dsets = list(it)
        assert len(dsets) == ladim_dset.dims['time'] + ladim_dset2.dims['time']

    def test_returns_correct_time_selection(self, ladim_dset):
        iterator = ladim_input.ladim_iterator([ladim_dset])
        particle_count = [d.particle_count.values.tolist() for d in iterator]
        assert particle_count == [[4, 4, 4, 4], [2, 2]]

        iterator = ladim_input.ladim_iterator([ladim_dset])
        time = [d.time.values for d in iterator]
        assert len(time) == 2
        assert time[0].astype(str).tolist() == ['2000-01-02T00:00:00.000000000'] * 4
        assert time[1].astype(str).tolist() == ['2000-01-03T00:00:00.000000000'] * 2

    def test_returns_correct_instance_selection(self, ladim_dset):
        iterator = ladim_input.ladim_iterator([ladim_dset])
        z = [d.Z.values.tolist() for d in iterator]
        assert z == [[0, 1, 2, 3], [4, 5]]

        iterator = ladim_input.ladim_iterator([ladim_dset])
        pid = [d.pid.values.tolist() for d in iterator]
        assert pid == [[0, 1, 2, 3], [1, 2]]

    def test_broadcasts_particle_variables(self, ladim_dset):
        iterator = ladim_input.ladim_iterator([ladim_dset])
        farm_id = [d.farm_id.values.tolist() for d in iterator]
        assert farm_id == [[12345, 12346, 12347, 12348], [12346, 12347]]

    def test_updates_instance_offset(self, ladim_dset):
        iterator = ladim_input.ladim_iterator([ladim_dset])
        offset = [d.instance_offset.values.tolist() for d in iterator]
        assert offset == [0, 4]

    def test_accepts_timestep_selection_when_multiple_datasets(self, ladim_dset, ladim_dset2):
        # No timestep selection
        chunks = list(ladim_input.ladim_iterator([ladim_dset, ladim_dset2]))
        assert len(chunks) == 4

        times = [c.time.values[0].astype('datetime64[D]').astype(str) for c in chunks]
        assert times == ['2000-01-02', '2000-01-03', '2000-01-04', '2000-01-05']

        # Select first and fourth time step
        chunks = list(ladim_input.ladim_iterator([ladim_dset, ladim_dset2], timesteps=[0, 3]))
        assert len(chunks) == 2

        times = [c.time.values[0].astype('datetime64[D]').astype(str) for c in chunks]
        assert times == ['2000-01-02', '2000-01-05']

    def test_disregards_nonunique_timestep_selectors(self, ladim_dset):
        # Select first and second timestep
        chunks_1 = list(ladim_input.ladim_iterator([ladim_dset], timesteps=[0, 1]))
        assert len(chunks_1) == 2

        # Select multiple timesteps: We still get 2 chunks
        chunks_2 = list(ladim_input.ladim_iterator([ladim_dset], timesteps=[1, 0, 1]))
        assert len(chunks_2) == 2

        # And the time values are ordered
        assert chunks_2[0].time.values[0] < chunks_2[1].time.values[0]


class Test_LadimInputStream_scan:
    def test_can_return_min_value(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        spec = dict(X=['min'])
        out = dset.scan(spec)
        assert out == dict(X=dict(min=5))

    def test_can_return_max_value(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        spec = dict(X=['max'])
        out = dset.scan(spec)
        assert out == dict(X=dict(max=6))

    def test_can_return_multiple_stats(self, ladim_dset, ladim_dset2):
        ladim_dset3 = ladim_dset2.copy(deep=True)
        ladim_dset3['X'] += 10
        ladim_dset3['Y'] += 10
        dset = ladim_input.LadimInputStream([ladim_dset, ladim_dset3])
        spec = dict(X=['max'], Y=['min', 'max'])
        out = dset.scan(spec)
        assert out == dict(X=dict(max=16), Y=dict(min=60, max=72))


class Test_LadimInputStream_assign:
    def test_can_add_numexpr_variables(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        dset.add_derived_variable('sumcoords', "X + Y")
        chunk = next(dset.chunks())
        assert chunk.sumcoords.values.tolist() == (chunk.X + chunk.Y).values.tolist()

    def test_can_add_agg_variable(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        dset.add_aggregation_variable('X', 'max')
        dset.add_aggregation_variable('Y', 'min')
        chunk = next(dset.chunks())
        assert chunk.MAX_X.values.tolist() == ladim_dset.X.max().values.tolist()
        assert chunk.MIN_Y.values.tolist() == ladim_dset.Y.min().values.tolist()

    def test_can_add_unique(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        dset.add_aggregation_variable('X', 'unique')
        unique_x = dset.get_aggregation_value('UNIQUE_X')
        assert unique_x.values.tolist() == np.unique(ladim_dset.X.values).tolist()


class Test_LadimInputStream:
    def test_can_initialise_from_xr_dataset(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        next(dset.chunks())

    def test_can_initialise_from_multiple_xr_datasets(self, ladim_dset, ladim_dset2):
        dset = ladim_input.LadimInputStream([ladim_dset, ladim_dset2])
        next(dset.chunks())

    def test_reads_one_timestep_at_the_time(self, ladim_dset, ladim_dset2):
        dset = ladim_input.LadimInputStream([ladim_dset, ladim_dset2])
        pids = list(c.pid.values.tolist() for c in dset.chunks())
        assert len(pids) == ladim_dset.dims['time'] + ladim_dset2.dims['time']
        assert pids == [[0, 1, 2, 3], [1, 2], [0, 1, 2, 3], [1, 2]]

    def test_broadcasts_time_vars_when_reading(self, ladim_dset, ladim_dset2):
        dset = ladim_input.LadimInputStream([ladim_dset, ladim_dset2])
        counts = list(c.particle_count.values.tolist() for c in dset.chunks())
        assert counts == [[4, 4, 4, 4], [2, 2], [4, 4, 4, 4], [2, 2]]

    def test_broadcasts_particle_vars_when_reading(self, ladim_dset, ladim_dset2):
        dset = ladim_input.LadimInputStream([ladim_dset, ladim_dset2])
        farmid = list(c.farm_id.values.tolist() for c in dset.chunks())
        assert farmid == [
            [12345, 12346, 12347, 12348],
            [12346, 12347],
            [12345, 12346, 12347, 12348],
            [12346, 12347],
        ]

    def test_can_apply_filter_string(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)

        # No filter: 6 particle instances
        chunks = xr.concat(dset.chunks(), dim='pid')
        assert chunks.dims['pid'] == 6

        # With filter: 4 particle instances
        filters = "farm_id != 12346"
        chunks = xr.concat(dset.chunks(filters=filters), dim='pid')
        assert chunks.dims['pid'] == 4

        # A more complex filter expression
        filters = "(farm_id > 12345) & (farm_id < 12347)"
        chunks = xr.concat(dset.chunks(filters=filters), dim='pid')
        assert chunks.dims['pid'] == 2

    def test_can_apply_particle_filter(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)

        # No filter: 6 particle instances
        chunks = xr.concat(dset.chunks(), dim='pid')
        assert chunks.dims['pid'] == 6

        # With filter: Keeps only the first time the condition is triggered
        filters = "Z >= 2"
        chunks = xr.concat(dset.chunks(particle_filter=filters), dim='pid')
        assert chunks['pid'].values.tolist() == [2, 3, 1]
        assert chunks['Z'].values.tolist() == [2, 3, 4]

        # Combination of particle filter and regular filter
        chunk_iter = dset.chunks(particle_filter="Z >= 2", filters="pid != 3")
        chunks = xr.concat(chunk_iter, dim='pid')
        assert chunks['pid'].values.tolist() == [2, 1]
        assert chunks['Z'].values.tolist() == [2, 4]

    def test_can_apply_timestep_filter(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)

        # No filter: 6 particle instances
        chunks = xr.concat(dset.chunks(), dim='pid')
        assert chunks.dims['pid'] == 6

        # Just first timestep: 4 particle instances
        chunks = xr.concat(dset.chunks(timestep_filter=[0]), dim='pid')
        assert chunks.dims['pid'] == 4

        # Just second timestep: 2 particle instances
        chunks = xr.concat(dset.chunks(timestep_filter=[1]), dim='pid')
        assert chunks.dims['pid'] == 2

    def test_can_add_weights_from_string_expression(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        dset.add_derived_variable('weights', 'X + Y')
        chunk = next(c for c in dset.chunks())
        assert 'weights' in chunk
        assert len(chunk['weights']) > 0
        assert chunk['weights'].values.tolist() == list(
            chunk['X'].values + chunk['Y'].values
        )

    def test_can_return_attributes_of_variables(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        assert dset.attributes['Z']['standard_name'] == 'depth'


class Test_LadimInputStream_grid:
    def test_can_add_grid_variables(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        grid = xr.DataArray(
            data=[100, 200, 300],
            coords=dict(Z=[0, 4, 8]),
            name="multiplier",
        )
        dset.add_grid_variable(grid, 'linear')
        chunk = next(dset.chunks())
        assert "multiplier" in chunk.variables
        assert chunk.multiplier.values.tolist() == [100, 125, 150, 175]

    def test_can_use_interp_method_nearest(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        grid = xr.DataArray(
            data=[100, 200, 300],
            coords=dict(Z=[0, 4, 8]),
            name="multiplier",
        )
        dset.add_grid_variable(grid, 'nearest')
        chunk = next(dset.chunks())
        assert "multiplier" in chunk.variables
        assert chunk.multiplier.values.tolist() == [100, 100, 100, 200]

    def test_can_use_interp_method_bin(self, ladim_dset):
        dset = ladim_input.LadimInputStream(ladim_dset)
        grid = xr.DataArray(
            data=[100, 200, 300],
            coords=dict(Z=[0, 4, 8]),
            name="multiplier",
        )
        dset.add_grid_variable(grid, 'bin')
        chunk = next(dset.chunks())
        assert "multiplier" in chunk.variables
        assert chunk.multiplier.values.tolist() == [100, 100, 100, 100]


class Test_update_agg:
    def test_can_compute_max(self):
        assert ladim_input.update_agg(None, 'max', [1, 2, 3]) == 3
        assert ladim_input.update_agg(4, 'max', [1, 2, 3]) == 4
        assert ladim_input.update_agg(2, 'max', [1, 2, 3]) == 3

    def test_can_compute_min(self):
        assert ladim_input.update_agg(None, 'min', [1, 2, 3]) == 1
        assert ladim_input.update_agg(0, 'min', [1, 2, 3]) == 0
        assert ladim_input.update_agg(2, 'min', [1, 2, 3]) == 1

    def test_can_compute_unique(self):
        assert ladim_input.update_agg(None, 'unique', [1, 1, 3]) == [1, 3]
        assert ladim_input.update_agg([], 'unique', [1, 1, 3]) == [1, 3]
        assert ladim_input.update_agg([2, 3, 4], 'unique', [1, 1, 3]) == [1, 2, 3, 4]

    def test_can_update_init_if_old_data_is_supplied(self):
        old_data = np.array([1, 0, 0, 0, 0])
        old_mask = (old_data > 0)
        new_data = np.array([5, 6, 7, 8])
        new_pid = np.array([0, 2, 0, 2])

        data, mask = ladim_input.update_agg(
            old=(old_data, old_mask), aggfun='init', data=(new_data, new_pid))

        assert data.tolist() == [1, 0, 6, 0, 0]
        assert mask.tolist() == (data > 0).tolist()

    def test_can_update_final_if_old_data_is_supplied(self):
        old_data = np.array([1, 0, 0, 0, 0])
        old_mask = (old_data > 0)
        new_data = np.array([3, 4, 5, 6])
        new_pid = np.array([0, 2, 0, 2])

        data, mask = ladim_input.update_agg(
            old=(old_data, old_mask), aggfun='final', data=(new_data, new_pid))

        assert data.tolist() == [5, 0, 6, 0, 0]
        assert mask.tolist() == (data > 0).tolist()

    def test_can_initialize_init(self):
        new_data = np.array([5, 6, 7, 8])
        new_pid = np.array([0, 2, 0, 2])

        data, mask = ladim_input.update_agg(
            old=None, aggfun='init', data=(new_data, new_pid))

        assert data.tolist() == [5, 0, 6]
        assert mask.tolist() == (data > 0).tolist()

    def test_can_initialize_final(self):
        new_data = np.array([5, 6, 7, 8])
        new_pid = np.array([0, 2, 0, 2])

        data, mask = ladim_input.update_agg(
            old=None, aggfun='final', data=(new_data, new_pid))

        assert data.tolist() == [7, 0, 8]
        assert mask.tolist() == (data > 0).tolist()


class Test_create_pfilter:
    def test_checks_if_variable_is_triggered(self):
        pfilter = ladim_input.create_pfilter(spec='age >= 2')

        chunk = xr.Dataset(
            data_vars=dict(
                age=xr.Variable('particle_instance', [0, 1, 2, 3, 4, 5]),
                pid=xr.Variable('particle_instance', [6, 7, 8, 9, 10, 12])
            ),
        )
        is_triggered = pfilter(chunk)

        # First two particles are not triggered, the remaining are
        assert is_triggered.values.tolist() == [0, 0, 1, 1, 1, 1]

    def test_counts_only_first_occurrence(self):
        pfilter = ladim_input.create_pfilter(spec='age >= 2')

        chunk = xr.Dataset(
            data_vars=dict(
                pid=xr.Variable('particle_instance', [0, 1, 2, 3, 4, 5])
            ),
        )
        chunk1 = chunk.assign(age=xr.Variable('particle_instance', [0, 0, 0, 0, 4, 5]))
        chunk2 = chunk.assign(age=xr.Variable('particle_instance', [2, 0, 0, 0, 6, 6]))
        chunk3 = chunk.assign(age=xr.Variable('particle_instance', [0, 2, 0, 0, 9, 9]))

        pfilter(chunk1)
        pfilter(chunk2)
        is_triggered = pfilter(chunk3)
        assert is_triggered.values.tolist() == [0, 1, 0, 0, 0, 0]

    def test_works_with_large_pids(self):
        pfilter = ladim_input.create_pfilter(spec='age >= 2')

        chunk = xr.Dataset(
            data_vars=dict(
                age=xr.Variable('particle_instance', [0, 1, 2]),
                pid=xr.Variable('particle_instance', [1001, 1002, 1003])
            ),
        )

        pfilter(chunk)


class Test_create_varfunc:
    @patch('ladim_aggregate.input.get_varfunc_from_funcstring')
    def test_calls_correct_function_when_singledot_spec(self, mock_class):
        _ = ladim_input.create_varfunc('numpy.sum')
        assert mock_class.call_count == 1

    @patch('ladim_aggregate.input.get_varfunc_from_funcstring')
    def test_calls_correct_function_when_doubledot_spec(self, mock_class):
        _ = ladim_input.create_varfunc('numpy.linalg.inv')
        assert mock_class.call_count == 1

    @patch('ladim_aggregate.input.get_varfunc_from_numexpr')
    def test_calls_correct_function_when_numexpr_with_ints(self, mock_class):
        _ = ladim_input.create_varfunc('myvar * 123')
        assert mock_class.call_count == 1

    @patch('ladim_aggregate.input.get_varfunc_from_numexpr')
    def test_calls_correct_function_when_varname(self, mock_class):
        _ = ladim_input.create_varfunc('my_var')
        assert mock_class.call_count == 1

    @patch('ladim_aggregate.input.get_varfunc_from_numexpr')
    def test_calls_correct_function_when_numexpr_with_floats(self, mock_class):
        _ = ladim_input.create_varfunc('myvar * 1.23')
        assert mock_class.call_count == 1



import subprocess
from ladim_aggregate.script import SCRIPT_NAME
from ladim_aggregate import script, examples
import pytest
import netCDF4 as nc
import xarray as xr
from uuid import uuid4


class Test_main:
    def test_prints_help_message_when_no_arguments(self, capsys):
        script.main()
        out = capsys.readouterr().out
        assert out.startswith('usage: ' + SCRIPT_NAME)

    def test_prints_help_message_when_help_argument(self, capsys):
        with pytest.raises(SystemExit):
            script.main('--help')
        out = capsys.readouterr().out
        assert out.startswith('usage: ' + SCRIPT_NAME)


named_examples_all = examples.Example.available()
named_examples = ['grid_2D']


class Test_command_line_script:
    @pytest.mark.parametrize("name", named_examples)
    def test_can_extract_and_run_example(self, name, tmp_path):
        import os
        os.chdir(tmp_path)
        r = subprocess.run([SCRIPT_NAME, '--example', name], stdout=subprocess.PIPE)
        assert r.stdout.decode('utf-8') == ''
        files = {f.name for f in tmp_path.glob('*')}
        assert 'aggregate.yaml' in files
        assert 'count.nc' in files


class Test_run_conf:
    @pytest.fixture
    def infile(self):
        dset = xr.Dataset(
            coords=dict(
                time=[100, 200],
            ),
            data_vars=dict(
                particle_count=xr.Variable('time', [1, 2]),
                x=xr.Variable('particle_instance', [7, 13, 15]),
                pid=xr.Variable('particle_instance', [0, 0, 1]),
            ),
        )
        return dset

    @pytest.fixture
    def outfile(self):
        with nc.Dataset(uuid4(), 'w', diskless=True) as dset:
            yield dset

    def test_works_with_in_memory_datasets(self, infile, outfile):
        conf = dict(
            bins=dict(x=[0, 10, 20]),
            infile=infile,
            outfile=outfile,
        )

        script.run_conf(conf)

        assert outfile.variables['histogram'][:].tolist() == [1, 2]



SCRIPT_NAME = "crecon"


def main_from_command_line():
    import sys
    main(*sys.argv[1:])


def main(*args):
    import argparse

    from .examples import Example
    available = Example.available()
    sort_order = [
        'grid_2D', 'grid_3D', 'time', 'filter', 'weights', 'wgt_tab', 'last', 'groupby',
        'multi', 'blur', 'crs', 'density', 'geotag', 'connect',
    ]
    example_names = [n for n in sort_order if n in available]
    example_names += [n for n in available if n not in example_names]

    # Planned for the future:
    # '  blur:     Apply a blurring filter the output grid\n'

    example_list = []
    for name in example_names:
        ex = Example(name)
        example_list.append(f'  {name:8}  {ex.descr}')

    from . import __version__ as version_str

    parser = argparse.ArgumentParser(
        prog='crecon',
        description=(
            f"CRECON - CREate CONcentration files (v. {version_str})\n\n"
            "This script converts LADiM particle files to netCDF\n"
            "concentration files.\n\n"
        ),
        epilog=(
            'The program includes several built-in examples:\n'
            + "\n".join(example_list) +
            '\n\nUse "crecon --example name_of_example" to run any of these.\n'
            'Example files and output files are extracted to the current directory.\n'
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        'config_file',
        help="File describing the aggregation options (YAML format)"
    )

    parser.add_argument(
        '--example',
        action='store_true',
        help="Run a built-in example"
    )

    # If called with too few arguments, print usage information
    if len(args) < 1:
        parser.print_help()
        return

    parsed_args = parser.parse_args(args)
    config_file = parsed_args.config_file

    import logging
    from . import __version__ as version_str
    init_logger()

    try:
        logger = logging.getLogger(__name__)
        logger.info(f'Starting CRECON, version {version_str}')

        # Extract example if requested
        if parsed_args.example:
            ex = Example(config_file)
            config_file = ex.extract()

        import yaml
        logger.info(f'Open config file "{config_file}"')
        with open(config_file, encoding='utf-8') as f:
            config = yaml.safe_load(f)

        run_conf(config)

    finally:
        close_logger()


def run_conf(config):
    """
    Run crecon simulation using a configuration dict

    :param config: Configuration
    """
    import logging
    logger = logging.getLogger(__name__)

    logger.debug(f'Input file pattern: "{config["infile"]}"')
    from .input import LadimInputStream

    dset_in = LadimInputStream(config['infile'])
    logger.debug(f'Number of input datasets: {len(dset_in.datasets)}')

    logger.info(f'Create output file "{config["outfile"]}"')
    from .output import MultiDataset

    with MultiDataset(config['outfile']) as dset_out:
        run(dset_in, config, dset_out)


def run(dset_in, config, dset_out, filedata=None):
    from .histogram import Histogrammer, autobins
    from .parseconfig import parse_config, load_config
    import numpy as np

    # Modify configuration dict by reformatting and appending default values
    config = parse_config(config)
    config = load_config(config, filedata)

    # Read some params
    filesplit_dims = config.get('filesplit_dims', [])
    filter_spec = config.get('filter', None)
    tsfilter_spec = config.get('filter_timestep', None)
    pfilter_spec = config.get('filter_particle', None)

    # Add geotagging
    if 'geotag' in config:
        for k in config['geotag']['attrs']:
            spec = ('geotag', dict(
                attribute=k,
                x_var=config['geotag']['coords']['x'],
                y_var=config['geotag']['coords']['y'],
                geojson=config['geotag']['geojson'],
                missing=config['geotag']['outside_value'],
            ))
            dset_in.add_derived_variable(varname=k, definition=spec)

    # Add grid variables
    for gridvar_spec in config.get('grid', []):
        dset_in.add_grid_variable(
            data_array=gridvar_spec['data'],
            method=gridvar_spec['method'],
        )

    # Add derived variables
    for derived_name, derived_spec in config.get('derived', dict()).items():
        dset_in.add_derived_variable(varname=derived_name, definition=derived_spec)

    # Add weights
    if 'weights' in config:
        dset_in.add_derived_variable(varname='_auto_weights', definition=config['weights'])

    # Prepare histogram bins
    bins = autobins(config['bins'], dset_in)
    hist = Histogrammer(bins=bins)
    coords = hist.coords

    # Create output coordinate variables
    for coord_name, coord_info in coords.items():
        dset_out.createCoord(
            varname=coord_name,
            data=coord_info['centers'],
            attrs=coord_info.get('attrs', dict()),
            cross_dataset=coord_name in filesplit_dims,
        )

    # Create aggregation variable
    hist_dtype = np.float32 if 'weights' in config else np.int32
    dset_out.createVariable(
        varname=config['output_varname'],
        data=np.array(0, dtype=hist_dtype),
        dims=tuple(coords.keys()),
    )

    # Add projection information
    if 'projection' in config:
        from .proj import write_projection
        write_projection(dset_out, config['projection'])

    import logging
    logger = logging.getLogger(__name__)

    # Read ladim file timestep by timestep
    dset_in_iterator = dset_in.chunks(
        filters=filter_spec,
        timestep_filter=tsfilter_spec,
        particle_filter=pfilter_spec,
    )

    for chunk_in in dset_in_iterator:
        if chunk_in.dims['pid'] == 0:
            continue

        # Write histogram values to file
        for chunk_out in hist.make(chunk_in):
            txt = ", ".join([f'{a.start}:{a.stop}' for a in chunk_out['indices']])
            logger.debug(f'Write output chunk [{txt}]')
            dset_out.incrementData(
                varname=config['output_varname'],
                data=chunk_out['values'],
                idx=chunk_out['indices'],
            )

    return dset_out


def init_logger(loglevel=None):
    import logging
    if loglevel is None:
        loglevel = logging.INFO

    package_name = str(__name__).split('.', maxsplit=1)[0]
    package_logger = logging.getLogger(package_name)
    package_logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s  %(name)s:%(levelname)s - %(message)s')

    ch = logging.StreamHandler()
    ch.setLevel(loglevel)
    ch.setFormatter(formatter)
    package_logger.addHandler(ch)


def close_logger():
    import logging
    package_name = str(__name__).split('.', maxsplit=1)[0]
    package_logger = logging.getLogger(package_name)

    # Close the log handlers
    handlers = [h for h in package_logger.handlers]  # Make a copy, otherwise the loop will fail
    for handler in handlers:
        handler.close()
        package_logger.removeHandler(handler)



from ladim_aggregate import parseconfig


class Test_parse_config:
    def test_output_equals_input_when_unknown_keywords(self):
        conf = dict(unknown_keyword="some_values")
        conf_out = parseconfig.parse_config(conf)
        assert conf_out["unknown_keyword"] == "some_values"

    def test_dont_change_bin_if_present_and_in_filesplit_dims(self):
        conf = dict(
            bins=dict(x=[1, 2, 3], farm_id=[12345, 23456]),
            filesplit_dims=['farm_id'],
        )
        conf_out = parseconfig.parse_config(conf)
        assert conf_out['bins']['farm_id'] == [12345, 23456]

    def test_adds_groupby_bin_on_top_if_missing_and_in_filesplit_dims(self):
        conf = dict(
            bins=dict(x=[1, 2, 3]),
            filesplit_dims=['farm_id'],
        )
        conf_out = parseconfig.parse_config(conf)
        assert conf_out['bins']['farm_id'] == 'group_by'
        assert list(conf_out['bins'].keys()) == ['farm_id', 'x']



import xarray as xr
from matplotlib.path import Path
import numpy as np


def create_geotagger(attribute, x_var, y_var, geojson, missing=np.nan):
    """
    Create geotagger function.

    :param attribute: Polygon attribute to return
    :param x_var: Name of x coordinate variable
    :param y_var: Name of y coordinate variable
    :param geojson: A dict representation of a geojson file
    :param missing: Value to return if coordinates does not match a polygon
    :return: A geotagger function
    """
    props = np.array([f['properties'][attribute] for f in geojson['features']] + [missing])
    coords = [
        np.asarray(f['geometry']['coordinates']).reshape((-1, 2))
        for f in geojson['features']
    ]
    paths = [Path(vertices=c) for c in coords]

    def geotagger(chunk):
        """
        Returns attributes based on geographical position

        :param chunk: An xarray dataset containing coordinates
        :return: An xarray variable with attributes from enclosing polygons
        """
        x = chunk[x_var].values
        y = chunk[y_var].values
        xy = np.stack([x, y]).T
        inside = np.asarray([p.contains_points(xy) for p in paths])
        first_nonzero = np.sum(np.cumsum(inside, axis=0) == 0, axis=0)
        return xr.Variable(dims='pid', data=props[first_nonzero])
    return geotagger



def parse_config(conf):
    """
    Parse an input configuration to ladim_aggregate, append default values and make
    formatting changes before it is passed onwards to the program.

    :param conf: Input configuration
    :return: Output configuration, with default values appended and formatting changes
    applied.
    """
    conf_out = conf.copy()

    conf_out['bins'] = conf.get('bins', dict())
    conf_out['grid'] = conf.get('grid', [])
    conf_out['output_varname'] = conf.get('output_varname', 'histogram')
    conf_out['filesplit_dims'] = conf.get('filesplit_dims', [])

    filesplit_bins = {k: "group_by" for k in conf_out['filesplit_dims']}
    conf_out['bins'] = {**filesplit_bins, **conf_out['bins']}

    if 'projection' in conf:
        conf_out['projection']['output_varname'] = conf_out['output_varname']

    for idx, item in enumerate(conf_out['grid']):
        if 'method' not in item:
            conf_out['grid'][idx]['method'] = 'linear'

    return conf_out


def load_config(config, filedata):
    import xarray as xr
    import logging
    logger = logging.getLogger(__name__)

    filedata = filedata or dict()

    # Load geotag file
    if 'geotag' in config:
        fname = config['geotag']['file']
        data = filedata.get(fname, None)
        if data is None:
            logger.info(f'Load geotag file "{fname}"')
            with open(fname, 'br') as f:
                data = f.read()

        import json
        config['geotag']['geojson'] = json.loads(data.decode(encoding='utf-8'))

    # Load grid files
    for grid_spec in config['grid']:
        fname = grid_spec['file']
        data = filedata.get(fname, None)  # type: xr.Dataset
        if data is None:
            logger.info(f'Load grid file "{grid_spec["file"]}"')
            with xr.open_dataset(grid_spec['file'], decode_cf=False) as data:
                grid_spec['data'] = data[grid_spec['variable']].copy(deep=True)
        else:
            grid_spec['data'] = data[grid_spec['variable']].copy(deep=True)

    return config
